from __future__ import annotations

from .base import MinerAlgoType
from .hashrate import EtHashHashRate
from .hashrate.unit import EtHashUnit


class EtHashAlgo(MinerAlgoType):
    hashrate: type[EtHashHashRate] = EtHashHashRate
    unit: type[EtHashUnit] = EtHashUnit

    name = "EtHash"



from pyasic.miners.backends import M2X
from pyasic.miners.device.models import M20SV10


class BTMinerM20SV10(M2X, M20SV10):
    pass


from pyasic.miners.device.models import M20SV20


class BTMinerM20SV20(M2X, M20SV20):
    pass


from pyasic.miners.device.models import M20SV30


class BTMinerM20SV30(M2X, M20SV30):
    pass



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from __future__ import annotations

import hashlib
import json
from pathlib import Path
from typing import Any

import aiofiles
import httpx

from pyasic import settings
from pyasic.errors import APIError
from pyasic.web.base import BaseWebAPI


class ePICWebAPI(BaseWebAPI):
    def __init__(self, ip: str) -> None:
        super().__init__(ip)
        self.username = "root"
        self.pwd = settings.get("default_epic_web_password", "letmein")
        self.port = 4028
        self.token = None

    async def send_command(
        self,
        command: str | bytes,
        ignore_errors: bool = False,
        allow_warning: bool = True,
        privileged: bool = False,
        **parameters: Any,
    ) -> dict:
        post = privileged or not parameters == {}

        async with httpx.AsyncClient(transport=settings.transport()) as client:
            for retry_cnt in range(settings.get("get_data_retries", 1)):
                try:
                    if parameters.get("form") is not None:
                        form_data = parameters["form"]
                        form_data.add_field("password", self.pwd)
                        response = await client.post(
                            f"http://{self.ip}:{self.port}/{command}",
                            timeout=5,
                            data=form_data,
                        )
                    if post:
                        response = await client.post(
                            f"http://{self.ip}:{self.port}/{command}",
                            timeout=5,
                            json={
                                **parameters,
                                "password": self.pwd,
                            },
                        )
                    else:
                        response = await client.get(
                            f"http://{self.ip}:{self.port}/{command}",
                            timeout=5,
                        )
                    if not response.status_code == 200:
                        if not ignore_errors:
                            raise APIError(
                                f"Web command {command} failed with status code {response.status_code}"
                            )
                        return {}
                    json_data = response.json()
                    if json_data:
                        # The API can return a fail status if the miner cannot return the requested data. Catch this and pass
                        if not json_data.get("result", True) and not post:
                            if retry_cnt < settings.get("get_data_retries", 1) - 1:
                                continue
                            if not ignore_errors:
                                raise APIError(json_data["error"])
                        return json_data
                    return {"success": True}
                except (httpx.HTTPError, json.JSONDecodeError, AttributeError):
                    pass

    async def multicommand(
        self, *commands: str, ignore_errors: bool = False, allow_warning: bool = True
    ) -> dict:
        data = {k: None for k in commands}
        data["multicommand"] = True
        for command in commands:
            data[command] = await self.send_command(command)
        return data

    async def restart_epic(self) -> dict:
        return await self.send_command("softreboot", privileged=True)

    async def reboot(self) -> dict:
        return await self.send_command("reboot", privileged=True)

    async def set_shutdown_temp(self, params: int) -> dict:
        return await self.send_command("shutdowntemp", param=params)

    async def set_critical_temp(self, params: int) -> dict:
        return await self.send_command("criticaltemp", param=params)

    async def set_fan(self, params: dict) -> dict:
        return await self.send_command("fanspeed", param=params)

    async def set_ptune_enable(self, params: bool) -> dict:
        return await self.send_command("perpetualtune", param=params)

    async def set_ptune_algo(self, params: dict) -> dict:
        return await self.send_command("perpetualtune/algo", param=params)

    async def set_pools(self, params: dict) -> dict:
        return await self.send_command("coin", param=params)

    async def pause_mining(self) -> dict:
        return await self.send_command("miner", param="Stop")

    async def resume_mining(self) -> dict:
        return await self.send_command("miner", param="Autostart")

    async def stop_mining(self) -> dict:
        return await self.send_command("miner", param="Stop")

    async def start_mining(self) -> dict:
        return await self.send_command("miner", param="Autostart")

    async def summary(self) -> dict:
        return await self.send_command("summary")

    async def hashrate(self) -> dict:
        return await self.send_command("hashrate")

    async def network(self) -> dict:
        return await self.send_command("network")

    async def capabilities(self) -> dict:
        return await self.send_command("capabilities")

    async def system_update(self, file: Path | str, keep_settings: bool = True):
        """Perform a system update by uploading a firmware file and sending a
        command to initiate the update."""

        # calculate the SHA256 checksum of the firmware file
        sha256_hash = hashlib.sha256()
        async with aiofiles.open(str(file), "rb") as f:
            while chunk := await f.read(8192):
                sha256_hash.update(chunk)
        checksum = sha256_hash.hexdigest()

        # prepare the multipart/form-data request
        form_data = aiohttp.FormData()
        form_data.add_field("checksum", checksum)
        form_data.add_field("keepsettings", str(keep_settings).lower())
        form_data.add_field("update.zip", open(file, "rb"), filename="update.zip")

        await self.send_command("systemupdate", form=form_data)



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M60SPlusVK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVK30

    expected_chips = 245
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60SPlusVK40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVK40

    expected_chips = 215
    expected_fans = 2
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M60SPlusVK50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVK50

    expected_chips = 225
    expected_fans = 2
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M60SPlusVK60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVK60

    expected_chips = 294
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60SPlusVK70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVK70

    expected_chips = 306
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60SPlusVL10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVL10

    expected_chips = 196
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60SPlusVL30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVL30

    expected_chips = 225
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60SPlusVL40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVL40

    expected_chips = 188
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60SPlusVL50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVL50

    expected_chips = 180
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60SPlusVL60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60SPlusVL60

    expected_chips = 172
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



from .M60 import (
    BTMinerM60VK6A,
    BTMinerM60VK10,
    BTMinerM60VK20,
    BTMinerM60VK30,
    BTMinerM60VK40,
    BTMinerM60VL10,
    BTMinerM60VL20,
    BTMinerM60VL30,
    BTMinerM60VL40,
    BTMinerM60VL50,
)
from .M60S import (
    BTMinerM60SVK10,
    BTMinerM60SVK20,
    BTMinerM60SVK30,
    BTMinerM60SVK40,
    BTMinerM60SVL10,
    BTMinerM60SVL20,
    BTMinerM60SVL30,
    BTMinerM60SVL40,
    BTMinerM60SVL50,
    BTMinerM60SVL60,
    BTMinerM60SVL70,
)
from .M60S_Plus import (
    BTMinerM60SPlusVK30,
    BTMinerM60SPlusVK40,
    BTMinerM60SPlusVK50,
    BTMinerM60SPlusVK60,
    BTMinerM60SPlusVK70,
    BTMinerM60SPlusVL10,
    BTMinerM60SPlusVL30,
    BTMinerM60SPlusVL40,
    BTMinerM60SPlusVL50,
    BTMinerM60SPlusVL60,
)
from .M60S_Plus_Plus import BTMinerM60SPlusPlusVL30, BTMinerM60SPlusPlusVL40
from .M61 import (
    BTMinerM61VK10,
    BTMinerM61VK20,
    BTMinerM61VK30,
    BTMinerM61VK40,
    BTMinerM61VL10,
    BTMinerM61VL30,
    BTMinerM61VL40,
    BTMinerM61VL50,
    BTMinerM61VL60,
)
from .M61S import BTMinerM61SVL10, BTMinerM61SVL20, BTMinerM61SVL30
from .M61S_Plus import BTMinerM61SPlusVL30
from .M62S_Plus import BTMinerM62SPlusVK30
from .M63 import (
    BTMinerM63VK10,
    BTMinerM63VK20,
    BTMinerM63VK30,
    BTMinerM63VL10,
    BTMinerM63VL30,
)
from .M63S import (
    BTMinerM63SVK10,
    BTMinerM63SVK20,
    BTMinerM63SVK30,
    BTMinerM63SVK60,
    BTMinerM63SVL10,
    BTMinerM63SVL50,
    BTMinerM63SVL60,
)
from .M63S_Plus import (
    BTMinerM63SPlusVK30,
    BTMinerM63SPlusVL10,
    BTMinerM63SPlusVL20,
    BTMinerM63SPlusVL30,
    BTMinerM63SPlusVL50,
)
from .M63S_Plus_Plus import BTMinerM63SPlusPlusVL20
from .M64 import BTMinerM64VL30, BTMinerM64VL40
from .M64S import BTMinerM64SVL30
from .M65S import BTMinerM65SVK20, BTMinerM65SVL60
from .M65S_Plus import BTMinerM65SPlusVK30
from .M66 import BTMinerM66VK20, BTMinerM66VK30, BTMinerM66VL20, BTMinerM66VL30
from .M66S import (
    BTMinerM66SVK20,
    BTMinerM66SVK30,
    BTMinerM66SVK40,
    BTMinerM66SVK50,
    BTMinerM66SVK60,
    BTMinerM66SVL10,
    BTMinerM66SVL20,
    BTMinerM66SVL30,
    BTMinerM66SVL40,
    BTMinerM66SVL50,
)
from .M66S_Plus import (
    BTMinerM66SPlusVK30,
    BTMinerM66SPlusVL10,
    BTMinerM66SPlusVL20,
    BTMinerM66SPlusVL30,
    BTMinerM66SPlusVL40,
    BTMinerM66SPlusVL60,
)
from .M66S_Plus_Plus import BTMinerM66SPlusPlusVL20
from .M67S import BTMinerM67SVK30



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M56SPlusVJ30


class BTMinerM56SPlusVJ30(M5X, M56SPlusVJ30):
    pass


from pyasic.miners.device.models import M56SPlusVK30


class BTMinerM56SPlusVK30(M5X, M56SPlusVK30):
    pass


from pyasic.miners.device.models import M56SPlusVK40


class BTMinerM56SPlusVK40(M5X, M56SPlusVK40):
    pass


from pyasic.miners.device.models import M56SPlusVK50


class BTMinerM56SPlusVK50(M5X, M56SPlusVK50):
    pass



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from typing import Optional

from pyasic import APIError
from pyasic.config import MinerConfig, MiningModeConfig
from pyasic.miners.backends import BMMiner
from pyasic.miners.data import (
    DataFunction,
    DataLocations,
    DataOptions,
    RPCAPICommand,
    WebAPICommand,
)
from pyasic.miners.device.firmware import HiveonFirmware
from pyasic.web.hiveon import HiveonWebAPI

HIVEON_MODERN_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_get_system_info", "get_system_info")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_get_system_info", "get_system_info")],
        ),
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [WebAPICommand("web_get_blink_status", "get_blink_status")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [WebAPICommand("web_get_conf", "get_miner_conf")],
        ),
    }
)


class HiveonModern(HiveonFirmware, BMMiner):
    data_locations = HIVEON_MODERN_DATA_LOC

    web: HiveonWebAPI
    _web_cls = HiveonWebAPI

    async def get_config(self) -> MinerConfig:
        data = await self.web.get_miner_conf()
        if data:
            self.config = MinerConfig.from_hiveon_modern(data)
        return self.config

    async def fault_light_on(self) -> bool:
        data = await self.web.blink(blink=True)
        if data:
            if data.get("code") == "B000":
                self.light = True
        return self.light

    async def fault_light_off(self) -> bool:
        data = await self.web.blink(blink=False)
        if data:
            if data.get("code") == "B100":
                self.light = False
        return self.light

    async def reboot(self) -> bool:
        data = await self.web.reboot()
        if data:
            return True
        return False

    async def stop_mining(self) -> bool:
        cfg = await self.get_config()
        cfg.mining_mode = MiningModeConfig.sleep()
        await self.send_config(cfg)
        return True

    async def resume_mining(self) -> bool:
        cfg = await self.get_config()
        cfg.mining_mode = MiningModeConfig.normal()
        await self.send_config(cfg)
        return True

    async def _get_wattage(self, rpc_stats: dict = None) -> Optional[int]:
        if not rpc_stats:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats:
            boards = rpc_stats.get("STATS")
            try:
                wattage_raw = boards[1]["chain_power"]
            except (KeyError, IndexError):
                pass
            else:
                # parse wattage position out of raw data
                return round(float(wattage_raw.split(" ")[0]))

    async def _get_hostname(self, web_get_system_info: dict = None) -> Optional[str]:
        if web_get_system_info is None:
            try:
                web_get_system_info = await self.web.get_system_info()
            except APIError:
                pass

        if web_get_system_info is not None:
            try:
                return web_get_system_info["hostname"]
            except KeyError:
                pass

    async def _get_mac(self, web_get_system_info: dict = None) -> Optional[str]:
        if web_get_system_info is None:
            try:
                web_get_system_info = await self.web.get_system_info()
            except APIError:
                pass

        if web_get_system_info is not None:
            try:
                return web_get_system_info["macaddr"]
            except KeyError:
                pass

        try:
            data = await self.web.get_network_info()
            if data:
                return data["macaddr"]
        except KeyError:
            pass

    async def _get_fault_light(
        self, web_get_blink_status: dict = None
    ) -> Optional[bool]:
        if self.light:
            return self.light

        if web_get_blink_status is None:
            try:
                web_get_blink_status = await self.web.get_blink_status()
            except APIError:
                pass

        if web_get_blink_status is not None:
            try:
                self.light = web_get_blink_status["blink"]
            except KeyError:
                pass
        return self.light

    async def _is_mining(self, web_get_conf: dict = None) -> Optional[bool]:
        if web_get_conf is None:
            try:
                web_get_conf = await self.web.get_miner_conf()
            except APIError:
                pass

        if web_get_conf is not None:
            try:
                if str(web_get_conf["bitmain-work-mode"]).isdigit():
                    return (
                        False if int(web_get_conf["bitmain-work-mode"]) == 1 else True
                    )
                return False
            except LookupError:
                pass


HIVEON_OLD_DATA_LOC = DataLocations(
    **{
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
    }
)


class HiveonOld(HiveonFirmware, BMMiner):
    data_locations = HIVEON_OLD_DATA_LOC

    async def _get_wattage(self, rpc_stats: dict = None) -> Optional[int]:
        if not rpc_stats:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats:
            boards = rpc_stats.get("STATS")
            try:
                wattage_raw = boards[1]["chain_power"]
            except (KeyError, IndexError):
                pass
            else:
                # parse wattage position out of raw data
                return round(float(wattage_raw.split(" ")[0]))



from typing import Optional

from pyasic import APIError
from pyasic.device.algorithm import AlgoHashRate
from pyasic.miners.backends import BMMiner
from pyasic.miners.data import (
    DataFunction,
    DataLocations,
    DataOptions,
    RPCAPICommand,
    WebAPICommand,
)
from pyasic.web.mskminer import MSKMinerWebAPI

MSKMINER_DATA_LOC = DataLocations(
    **{
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_info_v1", "info_v1")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
    }
)


class MSKMiner(BMMiner):
    """Handler for MSKMiner"""

    data_locations = MSKMINER_DATA_LOC

    web: MSKMinerWebAPI
    _web_cls = MSKMinerWebAPI

    async def _get_hashrate(self, rpc_stats: dict = None) -> Optional[AlgoHashRate]:
        # get hr from API
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                return self.algo.hashrate(
                    rate=float(rpc_stats["STATS"][0]["total_rate"]),
                    unit=self.algo.unit.GH,
                ).into(self.algo.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_wattage(self, rpc_stats: dict = None) -> Optional[int]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                return rpc_stats["STATS"][0]["total_power"]
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_mac(self, web_info_v1: dict = None) -> Optional[str]:
        if web_info_v1 is None:
            try:
                web_info_v1 = await self.web.info_v1()
            except APIError:
                pass

        if web_info_v1 is not None:
            try:
                return web_info_v1["network_info"]["result"]["macaddr"].upper()
            except (LookupError, ValueError, TypeError):
                pass



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

from pyasic.device.makes import MinerMake
from pyasic.miners.base import BaseMiner


class WhatsMinerMake(BaseMiner):
    make = MinerMake.WHATSMINER


class AntMinerMake(BaseMiner):
    make = MinerMake.ANTMINER


class AvalonMinerMake(BaseMiner):
    make = MinerMake.AVALONMINER


class InnosiliconMake(BaseMiner):
    make = MinerMake.INNOSILICON


class GoldshellMake(BaseMiner):
    make = MinerMake.GOLDSHELL


class AuradineMake(BaseMiner):
    make = MinerMake.AURADINE


class ePICMake(BaseMiner):
    make = MinerMake.EPIC


class BitAxeMake(BaseMiner):
    make = MinerMake.BITAXE


class LuckyMinerMake(BaseMiner):
    make = MinerMake.LUCKYMINER


class IceRiverMake(BaseMiner):
    make = MinerMake.ICERIVER


class HammerMake(BaseMiner):
    make = MinerMake.HAMMER


class VolcMinerMake(BaseMiner):
    make = MinerMake.VOLCMINER


class BraiinsMake(BaseMiner):
    make = MinerMake.BRAIINS


class ElphapexMake(BaseMiner):
    make = MinerMake.ELPHAPEX



from pyasic.miners.backends import M3X
from pyasic.miners.device.models import M33V10


class BTMinerM33V10(M3X, M33V10):
    pass


from pyasic.miners.device.models import M33V20


class BTMinerM33V20(M3X, M33V20):
    pass


from pyasic.miners.device.models import M33V30


class BTMinerM33V30(M3X, M33V30):
    pass



from .M50 import (
    M50VE30,
    M50VG30,
    M50VH10,
    M50VH20,
    M50VH30,
    M50VH40,
    M50VH50,
    M50VH60,
    M50VH70,
    M50VH80,
    M50VH90,
    M50VJ10,
    M50VJ20,
    M50VJ30,
    M50VJ40,
    M50VJ60,
    M50VK40,
    M50VK50,
)
from .M50S import (
    M50SVH10,
    M50SVH20,
    M50SVH30,
    M50SVH40,
    M50SVH50,
    M50SVJ10,
    M50SVJ20,
    M50SVJ30,
    M50SVJ40,
    M50SVJ50,
    M50SVK10,
    M50SVK20,
    M50SVK30,
    M50SVK50,
    M50SVK60,
    M50SVK70,
    M50SVK80,
    M50SVL20,
    M50SVL30,
)
from .M50S_Plus import (
    M50SPlusVH30,
    M50SPlusVH40,
    M50SPlusVJ30,
    M50SPlusVJ40,
    M50SPlusVJ60,
    M50SPlusVK10,
    M50SPlusVK20,
    M50SPlusVK30,
    M50SPlusVL10,
    M50SPlusVL20,
    M50SPlusVL30,
)
from .M50S_Plus_Plus import (
    M50SPlusPlusVK10,
    M50SPlusPlusVK20,
    M50SPlusPlusVK30,
    M50SPlusPlusVK40,
    M50SPlusPlusVK50,
    M50SPlusPlusVK60,
    M50SPlusPlusVL20,
    M50SPlusPlusVL30,
    M50SPlusPlusVL40,
    M50SPlusPlusVL50,
    M50SPlusPlusVL60,
)
from .M52S import M52SVK30
from .M52S_Plus_Plus import M52SPlusPlusVL10
from .M53 import M53VH30, M53VH40, M53VH50, M53VK30, M53VK60
from .M53H import M53HVH10
from .M53S import M53SVH20, M53SVH30, M53SVJ30, M53SVJ40, M53SVK30
from .M53S_Plus import M53SPlusVJ30, M53SPlusVJ40, M53SPlusVJ50, M53SPlusVK30
from .M53S_Plus_Plus import (
    M53SPlusPlusVK10,
    M53SPlusPlusVK20,
    M53SPlusPlusVK30,
    M53SPlusPlusVK50,
    M53SPlusPlusVL10,
    M53SPlusPlusVL30,
)
from .M54S_Plus_Plus import M54SPlusPlusVK30, M54SPlusPlusVL30, M54SPlusPlusVL40
from .M56 import M56VH30
from .M56S import M56SVH30, M56SVJ30, M56SVJ40
from .M56S_Plus import M56SPlusVJ30, M56SPlusVK30, M56SPlusVK40, M56SPlusVK50
from .M56S_Plus_Plus import (
    M56SPlusPlusVK10,
    M56SPlusPlusVK30,
    M56SPlusPlusVK40,
    M56SPlusPlusVK50,
)
from .M59 import M59VH30



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

from typing import List, Optional

from pyasic import MinerConfig
from pyasic.data import Fan, HashBoard
from pyasic.data.error_codes import MinerErrorData, X19Error
from pyasic.data.pools import PoolMetrics, PoolUrl
from pyasic.device.algorithm import AlgoHashRate
from pyasic.errors import APIError
from pyasic.miners.data import (
    DataFunction,
    DataLocations,
    DataOptions,
    RPCAPICommand,
    WebAPICommand,
)
from pyasic.miners.device.firmware import StockFirmware
from pyasic.rpc.ccminer import CCMinerRPCAPI
from pyasic.web.hammer import HammerWebAPI

HAMMER_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_get_system_info", "get_system_info")],
        ),
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_get_system_info", "get_system_info")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.ERRORS): DataFunction(
            "_get_errors",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [WebAPICommand("web_get_blink_status", "get_blink_status")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [WebAPICommand("web_get_conf", "get_miner_conf")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
    }
)


class BlackMiner(StockFirmware):
    """Handler for Hammer miners."""

    _rpc_cls = CCMinerRPCAPI
    rpc: CCMinerRPCAPI

    _web_cls = HammerWebAPI
    web: HammerWebAPI

    data_locations = HAMMER_DATA_LOC

    async def get_config(self) -> MinerConfig:
        data = await self.web.get_miner_conf()
        if data:
            self.config = MinerConfig.from_hammer(data)
        return self.config

    async def send_config(self, config: MinerConfig, user_suffix: str = None) -> None:
        self.config = config
        await self.web.set_miner_conf(config.as_hammer(user_suffix=user_suffix))

    async def fault_light_on(self) -> bool:
        data = await self.web.blink(blink=True)
        if data:
            if data.get("code") == "B000":
                self.light = True
        return self.light

    async def fault_light_off(self) -> bool:
        data = await self.web.blink(blink=False)
        if data:
            if data.get("code") == "B100":
                self.light = False
        return self.light

    async def reboot(self) -> bool:
        data = await self.web.reboot()
        if data:
            return True
        return False

    async def _get_api_ver(self, rpc_version: dict = None) -> Optional[str]:
        if rpc_version is None:
            try:
                rpc_version = await self.rpc.version()
            except APIError:
                pass

        if rpc_version is not None:
            try:
                self.api_ver = rpc_version["VERSION"][0]["API"]
            except LookupError:
                pass

        return self.api_ver

    async def _get_fw_ver(self, rpc_version: dict = None) -> Optional[str]:
        if rpc_version is None:
            try:
                rpc_version = await self.rpc.version()
            except APIError:
                pass

        if rpc_version is not None:
            try:
                self.fw_ver = rpc_version["VERSION"][0]["CompileTime"]
            except LookupError:
                pass

        return self.fw_ver

    async def _get_hashrate(self, rpc_summary: dict = None) -> Optional[AlgoHashRate]:
        # get hr from API
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return self.algo.hashrate(
                    rate=float(rpc_summary["SUMMARY"][0]["MHS 5s"]),
                    unit=self.algo.unit.MH,
                ).into(self.algo.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_hashboards(self, rpc_stats: dict = None) -> List[HashBoard]:
        hashboards = []

        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                board_offset = -1
                boards = rpc_stats["STATS"]
                if len(boards) > 1:
                    for board_num in range(1, 16, 5):
                        for _b_num in range(5):
                            b = boards[1].get(f"chain_acn{board_num + _b_num}")

                            if b and not b == 0 and board_offset == -1:
                                board_offset = board_num
                    if board_offset == -1:
                        board_offset = 1

                    real_slots = []

                    for i in range(board_offset, board_offset + 4):
                        try:
                            key = f"chain_acs{i}"
                            if boards[1].get(key, "") != "":
                                real_slots.append(i)
                        except LookupError:
                            pass

                    if len(real_slots) < 3:
                        real_slots = list(
                            range(board_offset, board_offset + self.expected_hashboards)
                        )

                    for i in real_slots:
                        hashboard = HashBoard(
                            slot=i - board_offset, expected_chips=self.expected_chips
                        )
                        temp = boards[1].get(f"temp{i}")
                        if temp:
                            hashboard.chip_temp = round(temp)
                            hashboard.temp = round(temp)

                        hashrate = boards[1].get(f"chain_rate{i}")
                        if hashrate:
                            hashboard.hashrate = self.algo.hashrate(
                                rate=float(hashrate), unit=self.algo.unit.MH
                            ).into(self.algo.unit.default)

                        chips = boards[1].get(f"chain_acn{i}")
                        if chips:
                            hashboard.chips = chips
                            hashboard.missing = False
                        if (not chips) or (not chips > 0):
                            hashboard.missing = True
                        hashboards.append(hashboard)
            except (LookupError, ValueError, TypeError):
                pass

        return hashboards

    async def _get_fans(self, rpc_stats: dict = None) -> List[Fan]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        fans = [Fan() for _ in range(self.expected_fans)]
        if rpc_stats is not None:
            try:
                fan_offset = -1

                for fan_num in range(1, 8, 4):
                    for _f_num in range(4):
                        f = rpc_stats["STATS"][1].get(f"fan{fan_num + _f_num}", 0)
                        if f and not f == 0 and fan_offset == -1:
                            fan_offset = fan_num
                if fan_offset == -1:
                    fan_offset = 1

                for fan in range(self.expected_fans):
                    fans[fan].speed = rpc_stats["STATS"][1].get(
                        f"fan{fan_offset+fan}", 0
                    )
            except LookupError:
                pass

        return fans

    async def _get_hostname(self, web_get_system_info: dict = None) -> Optional[str]:
        if web_get_system_info is None:
            try:
                web_get_system_info = await self.web.get_system_info()
            except APIError:
                pass

        if web_get_system_info is not None:
            try:
                return web_get_system_info["hostname"]
            except KeyError:
                pass

    async def _get_mac(self, web_get_system_info: dict = None) -> Optional[str]:
        if web_get_system_info is None:
            try:
                web_get_system_info = await self.web.get_system_info()
            except APIError:
                pass

        if web_get_system_info is not None:
            try:
                return web_get_system_info["macaddr"]
            except KeyError:
                pass

        try:
            data = await self.web.get_network_info()
            if data:
                return data["macaddr"]
        except KeyError:
            pass

    async def _get_errors(self, web_summary: dict = None) -> List[MinerErrorData]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        errors = []
        if web_summary is not None:
            try:
                for item in web_summary["SUMMARY"][0]["status"]:
                    try:
                        if not item["status"] == "s":
                            errors.append(X19Error(error_message=item["msg"]))
                    except KeyError:
                        continue
            except LookupError:
                pass
        return errors

    async def _get_fault_light(
        self, web_get_blink_status: dict = None
    ) -> Optional[bool]:
        if self.light:
            return self.light

        if web_get_blink_status is None:
            try:
                web_get_blink_status = await self.web.get_blink_status()
            except APIError:
                pass

        if web_get_blink_status is not None:
            try:
                self.light = web_get_blink_status["blink"]
            except KeyError:
                pass
        return self.light

    async def _get_expected_hashrate(
        self, rpc_stats: dict = None
    ) -> Optional[AlgoHashRate]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                expected_rate = rpc_stats["STATS"][1].get("total_rateideal")
                if expected_rate is None:
                    return self.sticker_hashrate.into(self.algo.unit.default)
                try:
                    rate_unit = rpc_stats["STATS"][1]["rate_unit"]
                except KeyError:
                    rate_unit = "MH"
                return self.algo.hashrate(
                    rate=float(expected_rate), unit=self.algo.unit.from_str(rate_unit)
                ).into(self.algo.unit.default)
            except LookupError:
                pass

    async def set_static_ip(
        self,
        ip: str,
        dns: str,
        gateway: str,
        subnet_mask: str = "255.255.255.0",
        hostname: str = None,
    ):
        if not hostname:
            hostname = await self.get_hostname()
        await self.web.set_network_conf(
            ip=ip,
            dns=dns,
            gateway=gateway,
            subnet_mask=subnet_mask,
            hostname=hostname,
            protocol=2,
        )

    async def set_dhcp(self, hostname: str = None):
        if not hostname:
            hostname = await self.get_hostname()
        await self.web.set_network_conf(
            ip="", dns="", gateway="", subnet_mask="", hostname=hostname, protocol=1
        )

    async def set_hostname(self, hostname: str):
        cfg = await self.web.get_network_info()
        dns = cfg["conf_dnsservers"]
        gateway = cfg["conf_gateway"]
        ip = cfg["conf_ipaddress"]
        subnet_mask = cfg["conf_netmask"]
        protocol = 1 if cfg["conf_nettype"] == "DHCP" else 2
        await self.web.set_network_conf(
            ip=ip,
            dns=dns,
            gateway=gateway,
            subnet_mask=subnet_mask,
            hostname=hostname,
            protocol=protocol,
        )

    async def _is_mining(self, web_get_conf: dict = None) -> Optional[bool]:
        if web_get_conf is None:
            try:
                web_get_conf = await self.web.get_miner_conf()
            except APIError:
                pass

        if web_get_conf is not None:
            try:
                if str(web_get_conf["bitmain-work-mode"]).isdigit():
                    return (
                        False if int(web_get_conf["bitmain-work-mode"]) == 1 else True
                    )
                return False
            except LookupError:
                pass

    async def _get_uptime(self, rpc_stats: dict = None) -> Optional[int]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                return int(rpc_stats["STATS"][1]["Elapsed"])
            except LookupError:
                pass

    async def _get_pools(self, rpc_pools: dict = None) -> List[PoolMetrics]:
        if rpc_pools is None:
            try:
                rpc_pools = await self.rpc.pools()
            except APIError:
                pass

        pools_data = []
        if rpc_pools is not None:
            try:
                pools = rpc_pools.get("POOLS", [])
                for pool_info in pools:
                    url = pool_info.get("URL")
                    pool_url = PoolUrl.from_str(url) if url else None
                    pool_data = PoolMetrics(
                        accepted=pool_info.get("Accepted"),
                        rejected=pool_info.get("Rejected"),
                        get_failures=pool_info.get("Get Failures"),
                        remote_failures=pool_info.get("Remote Failures"),
                        active=pool_info.get("Stratum Active"),
                        alive=pool_info.get("Status") == "Alive",
                        url=pool_url,
                        user=pool_info.get("User"),
                        index=pool_info.get("POOL"),
                    )
                    pools_data.append(pool_data)
            except LookupError:
                pass
        return pools_data



from pydantic import BaseModel


class BaseMinerError(BaseModel):
    @classmethod
    def fields(cls):
        return list(cls.model_fields.keys())

    def asdict(self) -> dict:
        return self.model_dump()

    def as_dict(self) -> dict:
        """Get this dataclass as a dictionary.

        Returns:
            A dictionary version of this class.
        """
        return self.asdict()



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M53SVH20


class BTMinerM53SVH20(M5X, M53SVH20):
    pass


from pyasic.miners.device.models import M53SVH30


class BTMinerM53SVH30(M5X, M53SVH30):
    pass


from pyasic.miners.device.models import M53SVJ30


class BTMinerM53SVJ30(M5X, M53SVJ30):
    pass


from pyasic.miners.device.models import M53SVJ40


class BTMinerM53SVJ40(M5X, M53SVJ40):
    pass


from pyasic.miners.device.models import M53SVK30


class BTMinerM53SVK30(M5X, M53SVK30):
    pass



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import enum
import ipaddress
import json
import re
import warnings
from typing import Any, AsyncGenerator, Callable

import anyio
import httpx

from pyasic import settings
from pyasic.logger import logger
from pyasic.miners.antminer import *
from pyasic.miners.auradine import *
from pyasic.miners.avalonminer import *
from pyasic.miners.backends import *
from pyasic.miners.base import AnyMiner
from pyasic.miners.bitaxe import *
from pyasic.miners.blockminer import *
from pyasic.miners.braiins import *
from pyasic.miners.device.makes import *
from pyasic.miners.elphapex import *
from pyasic.miners.goldshell import *
from pyasic.miners.hammer import *
from pyasic.miners.iceriver import *
from pyasic.miners.innosilicon import *
from pyasic.miners.luckyminer import *
from pyasic.miners.volcminer import *
from pyasic.miners.whatsminer import *


class MinerTypes(enum.Enum):
    ANTMINER = 0
    WHATSMINER = 1
    AVALONMINER = 2
    INNOSILICON = 3
    GOLDSHELL = 4
    BRAIINS_OS = 5
    VNISH = 6
    HIVEON = 7
    LUX_OS = 8
    EPIC = 9
    AURADINE = 10
    MARATHON = 11
    BITAXE = 12
    ICERIVER = 13
    HAMMER = 14
    VOLCMINER = 15
    LUCKYMINER = 16
    ELPHAPEX = 17
    MSKMINER = 18


MINER_CLASSES = {
    MinerTypes.ANTMINER: {
        None: type("AntminerUnknown", (BMMiner, AntMinerMake), {}),
        "ANTMINER D3": CGMinerD3,
        "ANTMINER HS3": BMMinerHS3,
        "ANTMINER L3+": BMMinerL3Plus,
        "ANTMINER KA3": BMMinerKA3,
        "ANTMINER KS3": BMMinerKS3,
        "ANTMINER DR5": CGMinerDR5,
        "ANTMINER KS5": BMMinerKS5,
        "ANTMINER KS5 PRO": BMMinerKS5Pro,
        "ANTMINER L7": BMMinerL7,
        "ANTMINER K7": BMMinerK7,
        "ANTMINER D7": BMMinerD7,
        "ANTMINER E9 PRO": BMMinerE9Pro,
        "ANTMINER D9": BMMinerD9,
        "ANTMINER S9": BMMinerS9,
        "ANTMINER S9I": BMMinerS9i,
        "ANTMINER S9J": BMMinerS9j,
        "ANTMINER T9": BMMinerT9,
        "ANTMINER L9": BMMinerL9,
        "ANTMINER Z15": CGMinerZ15,
        "ANTMINER Z15 PRO": BMMinerZ15Pro,
        "ANTMINER S17": BMMinerS17,
        "ANTMINER S17+": BMMinerS17Plus,
        "ANTMINER S17 PRO": BMMinerS17Pro,
        "ANTMINER S17E": BMMinerS17e,
        "ANTMINER T17": BMMinerT17,
        "ANTMINER T17+": BMMinerT17Plus,
        "ANTMINER T17E": BMMinerT17e,
        "ANTMINER S19": BMMinerS19,
        "ANTMINER S19L": BMMinerS19L,
        "ANTMINER S19 PRO": BMMinerS19Pro,
        "ANTMINER S19J": BMMinerS19j,
        "ANTMINER S19I": BMMinerS19i,
        "ANTMINER S19+": BMMinerS19Plus,
        "ANTMINER S19J88NOPIC": BMMinerS19jNoPIC,
        "ANTMINER S19PRO+": BMMinerS19ProPlus,
        "ANTMINER S19J PRO": BMMinerS19jPro,
        "ANTMINER S19 XP": BMMinerS19XP,
        "ANTMINER S19A": BMMinerS19a,
        "ANTMINER S19A PRO": BMMinerS19aPro,
        "ANTMINER S19 HYDRO": BMMinerS19Hydro,
        "ANTMINER S19 PRO HYD.": BMMinerS19ProHydro,
        "ANTMINER S19 PRO+ HYD.": BMMinerS19ProPlusHydro,
        "ANTMINER S19K PRO": BMMinerS19KPro,
        "ANTMINER S19J XP": BMMinerS19jXP,
        "ANTMINER T19": BMMinerT19,
        "ANTMINER BHB42XXX": BMMinerS19,
        "ANTMINER S21": BMMinerS21,
        "ANTMINER BHB68601": BMMinerS21,  # ???
        "ANTMINER BHB68606": BMMinerS21,  # ???
        "ANTMINER S21 PRO": BMMinerS21Pro,
        "ANTMINER T21": BMMinerT21,
        "ANTMINER S21 HYD.": BMMinerS21Hydro,
    },
    MinerTypes.WHATSMINER: {
        None: type("WhatsminerUnknown", (BTMiner, WhatsMinerMake), {}),
        "M20PV10": BTMinerM20PV10,
        "M20PV30": BTMinerM20PV30,
        "M20S+V30": BTMinerM20SPlusV30,
        "M20SV10": BTMinerM20SV10,
        "M20SV20": BTMinerM20SV20,
        "M20SV30": BTMinerM20SV30,
        "M20V10": BTMinerM20V10,
        "M21S+V20": BTMinerM21SPlusV20,
        "M21SV20": BTMinerM21SV20,
        "M21SV60": BTMinerM21SV60,
        "M21SV70": BTMinerM21SV70,
        "M21V10": BTMinerM21V10,
        "M29V10": BTMinerM29V10,
        "M30KV10": BTMinerM30KV10,
        "M30LV10": BTMinerM30LV10,
        "M30S++V10": BTMinerM30SPlusPlusV10,
        "M30S++V20": BTMinerM30SPlusPlusV20,
        "M30S++VE30": BTMinerM30SPlusPlusVE30,
        "M30S++VE40": BTMinerM30SPlusPlusVE40,
        "M30S++VE50": BTMinerM30SPlusPlusVE50,
        "M30S++VF40": BTMinerM30SPlusPlusVF40,
        "M30S++VG30": BTMinerM30SPlusPlusVG30,
        "M30S++VG40": BTMinerM30SPlusPlusVG40,
        "M30S++VG50": BTMinerM30SPlusPlusVG50,
        "M30S++VH10": BTMinerM30SPlusPlusVH10,
        "M30S++VH100": BTMinerM30SPlusPlusVH100,
        "M30S++VH110": BTMinerM30SPlusPlusVH110,
        "M30S++VH20": BTMinerM30SPlusPlusVH20,
        "M30S++VH30": BTMinerM30SPlusPlusVH30,
        "M30S++VH40": BTMinerM30SPlusPlusVH40,
        "M30S++VH50": BTMinerM30SPlusPlusVH50,
        "M30S++VH60": BTMinerM30SPlusPlusVH60,
        "M30S++VH70": BTMinerM30SPlusPlusVH70,
        "M30S++VH80": BTMinerM30SPlusPlusVH80,
        "M30S++VH90": BTMinerM30SPlusPlusVH90,
        "M30S++VI30": BTMinerM30SPlusPlusVI30,
        "M30S++VJ20": BTMinerM30SPlusPlusVJ20,
        "M30S++VJ30": BTMinerM30SPlusPlusVJ30,
        "M30S++VJ50": BTMinerM30SPlusPlusVJ50,
        "M30S++VJ60": BTMinerM30SPlusPlusVJ60,
        "M30S++VJ70": BTMinerM30SPlusPlusVJ70,
        "M30S++VK30": BTMinerM30SPlusPlusVK30,
        "M30S++VK40": BTMinerM30SPlusPlusVK40,
        "M30S+V10": BTMinerM30SPlusV10,
        "M30S+V100": BTMinerM30SPlusV100,
        "M30S+V20": BTMinerM30SPlusV20,
        "M30S+V30": BTMinerM30SPlusV30,
        "M30S+V40": BTMinerM30SPlusV40,
        "M30S+V50": BTMinerM30SPlusV50,
        "M30S+V60": BTMinerM30SPlusV60,
        "M30S+V70": BTMinerM30SPlusV70,
        "M30S+V80": BTMinerM30SPlusV80,
        "M30S+V90": BTMinerM30SPlusV90,
        "M30S+VE100": BTMinerM30SPlusVE100,
        "M30S+VE30": BTMinerM30SPlusVE30,
        "M30S+VE40": BTMinerM30SPlusVE40,
        "M30S+VE50": BTMinerM30SPlusVE50,
        "M30S+VE60": BTMinerM30SPlusVE60,
        "M30S+VE70": BTMinerM30SPlusVE70,
        "M30S+VE80": BTMinerM30SPlusVE80,
        "M30S+VE90": BTMinerM30SPlusVE90,
        "M30S+VF20": BTMinerM30SPlusVF20,
        "M30S+VF30": BTMinerM30SPlusVF30,
        "M30S+VG20": BTMinerM30SPlusVG20,
        "M30S+VG30": BTMinerM30SPlusVG30,
        "M30S+VG40": BTMinerM30SPlusVG40,
        "M30S+VG50": BTMinerM30SPlusVG50,
        "M30S+VG60": BTMinerM30SPlusVG60,
        "M30S+VH10": BTMinerM30SPlusVH10,
        "M30S+VH20": BTMinerM30SPlusVH20,
        "M30S+VH30": BTMinerM30SPlusVH30,
        "M30S+VH40": BTMinerM30SPlusVH40,
        "M30S+VH50": BTMinerM30SPlusVH50,
        "M30S+VH60": BTMinerM30SPlusVH60,
        "M30S+VH70": BTMinerM30SPlusVH70,
        "M30S+VI30": BTMinerM30SPlusVI30,
        "M30S+VJ30": BTMinerM30SPlusVJ30,
        "M30S+VJ40": BTMinerM30SPlusVJ40,
        "M30SV10": BTMinerM30SV10,
        "M30SV20": BTMinerM30SV20,
        "M30SV30": BTMinerM30SV30,
        "M30SV40": BTMinerM30SV40,
        "M30SV50": BTMinerM30SV50,
        "M30SV60": BTMinerM30SV60,
        "M30SV70": BTMinerM30SV70,
        "M30SV80": BTMinerM30SV80,
        "M30SVE10": BTMinerM30SVE10,
        "M30SVE20": BTMinerM30SVE20,
        "M30SVE30": BTMinerM30SVE30,
        "M30SVE40": BTMinerM30SVE40,
        "M30SVE50": BTMinerM30SVE50,
        "M30SVE60": BTMinerM30SVE60,
        "M30SVE70": BTMinerM30SVE70,
        "M30SVF10": BTMinerM30SVF10,
        "M30SVF20": BTMinerM30SVF20,
        "M30SVF30": BTMinerM30SVF30,
        "M30SVG10": BTMinerM30SVG10,
        "M30SVG20": BTMinerM30SVG20,
        "M30SVG30": BTMinerM30SVG30,
        "M30SVG40": BTMinerM30SVG40,
        "M30SVH10": BTMinerM30SVH10,
        "M30SVH20": BTMinerM30SVH20,
        "M30SVH30": BTMinerM30SVH30,
        "M30SVH40": BTMinerM30SVH40,
        "M30SVH50": BTMinerM30SVH50,
        "M30SVH60": BTMinerM30SVH60,
        "M30SVI20": BTMinerM30SVI20,
        "M30SVJ30": BTMinerM30SVJ30,
        "M30V10": BTMinerM30V10,
        "M30V20": BTMinerM30V20,
        "M31HV10": BTMinerM31HV10,
        "M31HV40": BTMinerM31HV40,
        "M31LV10": BTMinerM31LV10,
        "M31S+V10": BTMinerM31SPlusV10,
        "M31S+V100": BTMinerM31SPlusV100,
        "M31S+V20": BTMinerM31SPlusV20,
        "M31S+V30": BTMinerM31SPlusV30,
        "M31S+V40": BTMinerM31SPlusV40,
        "M31S+V50": BTMinerM31SPlusV50,
        "M31S+V60": BTMinerM31SPlusV60,
        "M31S+V80": BTMinerM31SPlusV80,
        "M31S+V90": BTMinerM31SPlusV90,
        "M31S+VE10": BTMinerM31SPlusVE10,
        "M31S+VE20": BTMinerM31SPlusVE20,
        "M31S+VE30": BTMinerM31SPlusVE30,
        "M31S+VE40": BTMinerM31SPlusVE40,
        "M31S+VE50": BTMinerM31SPlusVE50,
        "M31S+VE60": BTMinerM31SPlusVE60,
        "M31S+VE80": BTMinerM31SPlusVE80,
        "M31S+VF20": BTMinerM31SPlusVF20,
        "M31S+VF30": BTMinerM31SPlusVF30,
        "M31S+VG20": BTMinerM31SPlusVG20,
        "M31S+VG30": BTMinerM31SPlusVG30,
        "M31SEV10": BTMinerM31SEV10,
        "M31SEV20": BTMinerM31SEV20,
        "M31SEV30": BTMinerM31SEV30,
        "M31SV10": BTMinerM31SV10,
        "M31SV20": BTMinerM31SV20,
        "M31SV30": BTMinerM31SV30,
        "M31SV40": BTMinerM31SV40,
        "M31SV50": BTMinerM31SV50,
        "M31SV60": BTMinerM31SV60,
        "M31SV70": BTMinerM31SV70,
        "M31SV80": BTMinerM31SV80,
        "M31SV90": BTMinerM31SV90,
        "M31SVE10": BTMinerM31SVE10,
        "M31SVE20": BTMinerM31SVE20,
        "M31SVE30": BTMinerM31SVE30,
        "M31V10": BTMinerM31V10,
        "M31V20": BTMinerM31V20,
        "M32V10": BTMinerM32V10,
        "M32V20": BTMinerM32V20,
        "M33S++VG40": BTMinerM33SPlusPlusVG40,
        "M33S++VH20": BTMinerM33SPlusPlusVH20,
        "M33S++VH30": BTMinerM33SPlusPlusVH30,
        "M33S+VG20": BTMinerM33SPlusVG20,
        "M33S+VG30": BTMinerM33SPlusVG30,
        "M33S+VH20": BTMinerM33SPlusVH20,
        "M33S+VH30": BTMinerM33SPlusVH30,
        "M33SVG30": BTMinerM33SVG30,
        "M33V10": BTMinerM33V10,
        "M33V20": BTMinerM33V20,
        "M33V30": BTMinerM33V30,
        "M34S+VE10": BTMinerM34SPlusVE10,
        "M36S++VH30": BTMinerM36SPlusPlusVH30,
        "M36S+VG30": BTMinerM36SPlusVG30,
        "M36SVE10": BTMinerM36SVE10,
        "M39V10": BTMinerM39V10,
        "M39V20": BTMinerM39V20,
        "M39V30": BTMinerM39V30,
        "M50S++VK10": BTMinerM50SPlusPlusVK10,
        "M50S++VK20": BTMinerM50SPlusPlusVK20,
        "M50S++VK30": BTMinerM50SPlusPlusVK30,
        "M50S++VK40": BTMinerM50SPlusPlusVK40,
        "M50S++VK50": BTMinerM50SPlusPlusVK50,
        "M50S++VK60": BTMinerM50SPlusPlusVK60,
        "M50S++VL20": BTMinerM50SPlusPlusVL20,
        "M50S++VL30": BTMinerM50SPlusPlusVL30,
        "M50S++VL40": BTMinerM50SPlusPlusVL40,
        "M50S++VL50": BTMinerM50SPlusPlusVL50,
        "M50S++VL60": BTMinerM50SPlusPlusVL60,
        "M50S+VH30": BTMinerM50SPlusVH30,
        "M50S+VH40": BTMinerM50SPlusVH40,
        "M50S+VJ30": BTMinerM50SPlusVJ30,
        "M50S+VJ40": BTMinerM50SPlusVJ40,
        "M50S+VJ60": BTMinerM50SPlusVJ60,
        "M50S+VK10": BTMinerM50SPlusVK10,
        "M50S+VK20": BTMinerM50SPlusVK20,
        "M50S+VK30": BTMinerM50SPlusVK30,
        "M50S+VL10": BTMinerM50SPlusVL10,
        "M50S+VL20": BTMinerM50SPlusVL20,
        "M50S+VL30": BTMinerM50SPlusVL30,
        "M50SVH10": BTMinerM50SVH10,
        "M50SVH20": BTMinerM50SVH20,
        "M50SVH30": BTMinerM50SVH30,
        "M50SVH40": BTMinerM50SVH40,
        "M50SVH50": BTMinerM50SVH50,
        "M50SVJ10": BTMinerM50SVJ10,
        "M50SVJ20": BTMinerM50SVJ20,
        "M50SVJ30": BTMinerM50SVJ30,
        "M50SVJ40": BTMinerM50SVJ40,
        "M50SVJ50": BTMinerM50SVJ50,
        "M50SVK10": BTMinerM50SVK10,
        "M50SVK20": BTMinerM50SVK20,
        "M50SVK30": BTMinerM50SVK30,
        "M50SVK50": BTMinerM50SVK50,
        "M50SVK60": BTMinerM50SVK60,
        "M50SVK70": BTMinerM50SVK70,
        "M50SVK80": BTMinerM50SVK80,
        "M50SVL20": BTMinerM50SVL20,
        "M50SVL30": BTMinerM50SVL30,
        "M50VE30": BTMinerM50VE30,
        "M50VG30": BTMinerM50VG30,
        "M50VH10": BTMinerM50VH10,
        "M50VH20": BTMinerM50VH20,
        "M50VH30": BTMinerM50VH30,
        "M50VH40": BTMinerM50VH40,
        "M50VH50": BTMinerM50VH50,
        "M50VH60": BTMinerM50VH60,
        "M50VH70": BTMinerM50VH70,
        "M50VH80": BTMinerM50VH80,
        "M50VH90": BTMinerM50VH90,
        "M50VJ10": BTMinerM50VJ10,
        "M50VJ20": BTMinerM50VJ20,
        "M50VJ30": BTMinerM50VJ30,
        "M50VJ40": BTMinerM50VJ40,
        "M50VJ60": BTMinerM50VJ60,
        "M50VK40": BTMinerM50VK40,
        "M50VK50": BTMinerM50VK50,
        "M52S++VL10": BTMinerM52SPlusPlusVL10,
        "M52SVK30": BTMinerM52SVK30,
        "M53HVH10": BTMinerM53HVH10,
        "M53S++VK10": BTMinerM53SPlusPlusVK10,
        "M53S++VK20": BTMinerM53SPlusPlusVK20,
        "M53S++VK30": BTMinerM53SPlusPlusVK30,
        "M53S++VK50": BTMinerM53SPlusPlusVK50,
        "M53S++VL10": BTMinerM53SPlusPlusVL10,
        "M53S++VL30": BTMinerM53SPlusPlusVL30,
        "M53S+VJ30": BTMinerM53SPlusVJ30,
        "M53S+VJ40": BTMinerM53SPlusVJ40,
        "M53S+VJ50": BTMinerM53SPlusVJ50,
        "M53S+VK30": BTMinerM53SPlusVK30,
        "M53SVH20": BTMinerM53SVH20,
        "M53SVH30": BTMinerM53SVH30,
        "M53SVJ30": BTMinerM53SVJ30,
        "M53SVJ40": BTMinerM53SVJ40,
        "M53SVK30": BTMinerM53SVK30,
        "M53VH30": BTMinerM53VH30,
        "M53VH40": BTMinerM53VH40,
        "M53VH50": BTMinerM53VH50,
        "M53VK30": BTMinerM53VK30,
        "M53VK60": BTMinerM53VK60,
        "M54S++VK30": BTMinerM54SPlusPlusVK30,
        "M54S++VL30": BTMinerM54SPlusPlusVL30,
        "M54S++VL40": BTMinerM54SPlusPlusVL40,
        "M56S++VK10": BTMinerM56SPlusPlusVK10,
        "M56S++VK30": BTMinerM56SPlusPlusVK30,
        "M56S++VK40": BTMinerM56SPlusPlusVK40,
        "M56S++VK50": BTMinerM56SPlusPlusVK50,
        "M56S+VJ30": BTMinerM56SPlusVJ30,
        "M56S+VK30": BTMinerM56SPlusVK30,
        "M56S+VK40": BTMinerM56SPlusVK40,
        "M56S+VK50": BTMinerM56SPlusVK50,
        "M56SVH30": BTMinerM56SVH30,
        "M56SVJ30": BTMinerM56SVJ30,
        "M56SVJ40": BTMinerM56SVJ40,
        "M56VH30": BTMinerM56VH30,
        "M59VH30": BTMinerM59VH30,
        "M60S++VL30": BTMinerM60SPlusPlusVL30,
        "M60S++VL40": BTMinerM60SPlusPlusVL40,
        "M60S+VK30": BTMinerM60SPlusVK30,
        "M60S+VK40": BTMinerM60SPlusVK40,
        "M60S+VK50": BTMinerM60SPlusVK50,
        "M60S+VK60": BTMinerM60SPlusVK60,
        "M60S+VK70": BTMinerM60SPlusVK70,
        "M60S+VL10": BTMinerM60SPlusVL10,
        "M60S+VL30": BTMinerM60SPlusVL30,
        "M60S+VL40": BTMinerM60SPlusVL40,
        "M60S+VL50": BTMinerM60SPlusVL50,
        "M60S+VL60": BTMinerM60SPlusVL60,
        "M60SVK10": BTMinerM60SVK10,
        "M60SVK20": BTMinerM60SVK20,
        "M60SVK30": BTMinerM60SVK30,
        "M60SVK40": BTMinerM60SVK40,
        "M60SVL10": BTMinerM60SVL10,
        "M60SVL20": BTMinerM60SVL20,
        "M60SVL30": BTMinerM60SVL30,
        "M60SVL40": BTMinerM60SVL40,
        "M60SVL50": BTMinerM60SVL50,
        "M60SVL60": BTMinerM60SVL60,
        "M60SVL70": BTMinerM60SVL70,
        "M60VK10": BTMinerM60VK10,
        "M60VK20": BTMinerM60VK20,
        "M60VK30": BTMinerM60VK30,
        "M60VK40": BTMinerM60VK40,
        "M60VK6A": BTMinerM60VK6A,
        "M60VL10": BTMinerM60VL10,
        "M60VL20": BTMinerM60VL20,
        "M60VL30": BTMinerM60VL30,
        "M60VL40": BTMinerM60VL40,
        "M60VL50": BTMinerM60VL50,
        "M61S+VL30": BTMinerM61SPlusVL30,
        "M61SVL10": BTMinerM61SVL10,
        "M61SVL20": BTMinerM61SVL20,
        "M61SVL30": BTMinerM61SVL30,
        "M61VK10": BTMinerM61VK10,
        "M61VK20": BTMinerM61VK20,
        "M61VK30": BTMinerM61VK30,
        "M61VK40": BTMinerM61VK40,
        "M61VL10": BTMinerM61VL10,
        "M61VL30": BTMinerM61VL30,
        "M61VL40": BTMinerM61VL40,
        "M61VL50": BTMinerM61VL50,
        "M61VL60": BTMinerM61VL60,
        "M62S+VK30": BTMinerM62SPlusVK30,
        "M63S++VL20": BTMinerM63SPlusPlusVL20,
        "M63S+VK30": BTMinerM63SPlusVK30,
        "M63S+VL10": BTMinerM63SPlusVL10,
        "M63S+VL20": BTMinerM63SPlusVL20,
        "M63S+VL30": BTMinerM63SPlusVL30,
        "M63S+VL50": BTMinerM63SPlusVL50,
        "M63SVK10": BTMinerM63SVK10,
        "M63SVK20": BTMinerM63SVK20,
        "M63SVK30": BTMinerM63SVK30,
        "M63SVK60": BTMinerM63SVK60,
        "M63SVL10": BTMinerM63SVL10,
        "M63SVL50": BTMinerM63SVL50,
        "M63SVL60": BTMinerM63SVL60,
        "M63VK10": BTMinerM63VK10,
        "M63VK20": BTMinerM63VK20,
        "M63VK30": BTMinerM63VK30,
        "M63VL10": BTMinerM63VL10,
        "M63VL30": BTMinerM63VL30,
        "M64SVL30": BTMinerM64SVL30,
        "M64VL30": BTMinerM64VL30,
        "M64VL40": BTMinerM64VL40,
        "M65S+VK30": BTMinerM65SPlusVK30,
        "M65SVK20": BTMinerM65SVK20,
        "M65SVL60": BTMinerM65SVL60,
        "M66S++VL20": BTMinerM66SPlusPlusVL20,
        "M66S+VK30": BTMinerM66SPlusVK30,
        "M66S+VL10": BTMinerM66SPlusVL10,
        "M66S+VL20": BTMinerM66SPlusVL20,
        "M66S+VL30": BTMinerM66SPlusVL30,
        "M66S+VL40": BTMinerM66SPlusVL40,
        "M66S+VL60": BTMinerM66SPlusVL60,
        "M66SVK20": BTMinerM66SVK20,
        "M66SVK30": BTMinerM66SVK30,
        "M66SVK40": BTMinerM66SVK40,
        "M66SVK50": BTMinerM66SVK50,
        "M66SVK60": BTMinerM66SVK60,
        "M66SVL10": BTMinerM66SVL10,
        "M66SVL20": BTMinerM66SVL20,
        "M66SVL30": BTMinerM66SVL30,
        "M66SVL40": BTMinerM66SVL40,
        "M66SVL50": BTMinerM66SVL50,
        "M66VK20": BTMinerM66VK20,
        "M66VK30": BTMinerM66VK30,
        "M66VL20": BTMinerM66VL20,
        "M66VL30": BTMinerM66VL30,
        "M67SVK30": BTMinerM67SVK30,
        "M70VM30": BTMinerM70VM30,
    },
    MinerTypes.AVALONMINER: {
        None: type("AvalonUnknown", (AvalonMiner, AvalonMinerMake), {}),
        "AVALONMINER 721": CGMinerAvalon721,
        "AVALONMINER 741": CGMinerAvalon741,
        "AVALONMINER 761": CGMinerAvalon761,
        "AVALONMINER 821": CGMinerAvalon821,
        "AVALONMINER 841": CGMinerAvalon841,
        "AVALONMINER 851": CGMinerAvalon851,
        "AVALONMINER 921": CGMinerAvalon921,
        "AVALONMINER 1026": CGMinerAvalon1026,
        "AVALONMINER 1047": CGMinerAvalon1047,
        "AVALONMINER 1066": CGMinerAvalon1066,
        "AVALONMINER 1126PRO": CGMinerAvalon1126Pro,
        "AVALONMINER 1166PRO": CGMinerAvalon1166Pro,
        "AVALONMINER 1246": CGMinerAvalon1246,
        "AVALONMINER NANO3": CGMinerAvalonNano3,
        "AVALONMINER 15-194": CGMinerAvalon1566,
    },
    MinerTypes.INNOSILICON: {
        None: type("InnosiliconUnknown", (Innosilicon, InnosiliconMake), {}),
        "T3H+": InnosiliconT3HPlus,
        "A10X": InnosiliconA10X,
        "A11": InnosiliconA11,
        "A11MX": InnosiliconA11MX,
    },
    MinerTypes.GOLDSHELL: {
        None: type("GoldshellUnknown", (GoldshellMiner, GoldshellMake), {}),
        "GOLDSHELL CK5": GoldshellCK5,
        "GOLDSHELL HS5": GoldshellHS5,
        "GOLDSHELL KD5": GoldshellKD5,
        "GOLDSHELL KDMAX": GoldshellKDMax,
        "GOLDSHELL KDBOXII": GoldshellKDBoxII,
        "GOLDSHELL KDBOXPRO": GoldshellKDBoxPro,
    },
    MinerTypes.BRAIINS_OS: {
        None: BOSMiner,
        "ANTMINER S9": BOSMinerS9,
        "ANTMINER S17": BOSMinerS17,
        "ANTMINER S17+": BOSMinerS17Plus,
        "ANTMINER S17 PRO": BOSMinerS17Pro,
        "ANTMINER S17E": BOSMinerS17e,
        "ANTMINER T17": BOSMinerT17,
        "ANTMINER T17+": BOSMinerT17Plus,
        "ANTMINER T17E": BOSMinerT17e,
        "ANTMINER S19": BOSMinerS19,
        "ANTMINER S19+": BOSMinerS19Plus,
        "ANTMINER S19 PRO": BOSMinerS19Pro,
        "ANTMINER S19A": BOSMinerS19a,
        "ANTMINER S19A Pro": BOSMinerS19aPro,
        "ANTMINER S19J": BOSMinerS19j,
        "ANTMINER S19J88NOPIC": BOSMinerS19jNoPIC,
        "ANTMINER S19J PRO": BOSMinerS19jPro,
        "ANTMINER S19J PRO NOPIC": BOSMinerS19jProNoPIC,
        "ANTMINER S19J PRO+": BOSMinerS19jProPlus,
        "ANTMINER S19J PRO PLUS": BOSMinerS19jProPlus,
        "ANTMINER S19J PRO PLUS NOPIC": BOSMinerS19jProPlusNoPIC,
        "ANTMINER S19K PRO NOPIC": BOSMinerS19kProNoPIC,
        "ANTMINER S19K PRO": BOSMinerS19kProNoPIC,
        "ANTMINER S19 XP": BOSMinerS19XP,
        "ANTMINER S19 PRO+ HYD.": BOSMinerS19ProPlusHydro,
        "ANTMINER T19": BOSMinerT19,
        "ANTMINER S21": BOSMinerS21,
        "ANTMINER S21 PRO": BOSMinerS21Pro,
        "ANTMINER T21": BOSMinerT21,
        "BRAIINS MINI MINER BMM 100": BraiinsBMM100,
        "BRAIINS MINI MINER BMM 101": BraiinsBMM101,
    },
    MinerTypes.VNISH: {
        None: VNish,
        "L3+": VnishL3Plus,
        "ANTMINER L3+": VnishL3Plus,
        "ANTMINER L7": VnishL7,
        "ANTMINER S17+": VNishS17Plus,
        "ANTMINER S17 PRO": VNishS17Pro,
        "ANTMINER S19": VNishS19,
        "ANTMINER S19NOPIC": VNishS19NoPIC,
        "ANTMINER S19 PRO": VNishS19Pro,
        "ANTMINER S19J": VNishS19j,
        "ANTMINER S19I": VNishS19i,
        "ANTMINER S19J PRO": VNishS19jPro,
        "ANTMINER S19J PRO A": VNishS19jPro,
        "ANTMINER S19J PRO BB": VNishS19jPro,
        "ANTMINER S19J PRO+": VNishS19jPro,
        "ANTMINER S19A": VNishS19a,
        "ANTMINER S19A PRO": VNishS19aPro,
        "ANTMINER S19 PRO HYD.": VNishS19ProHydro,
        "ANTMINER S19K PRO": VNishS19kPro,
        "ANTMINER T19": VNishT19,
        "ANTMINER T21": VNishT21,
        "ANTMINER S21": VNishS21,
    },
    MinerTypes.EPIC: {
        None: ePIC,
        "ANTMINER S19": ePICS19,
        "ANTMINER S19 PRO": ePICS19Pro,
        "ANTMINER S19J": ePICS19j,
        "ANTMINER S19J PRO": ePICS19jPro,
        "ANTMINER S19J PRO+": ePICS19jProPlus,
        "ANTMINER S19K PRO": ePICS19kPro,
        "ANTMINER S19 XP": ePICS19XP,
        "ANTMINER S21": ePICS21,
        "ANTMINER S21 PRO": ePICS21Pro,
        "ANTMINER T21": ePICT21,
        "ANTMINER S19J PRO DUAL": ePICS19jProDual,
        "BLOCKMINER 520I": ePICBlockMiner520i,
        "BLOCKMINER 720I": ePICBlockMiner720i,
    },
    MinerTypes.HIVEON: {
        None: HiveonModern,
        "ANTMINER T9": HiveonT9,
        "ANTMINER S19JPRO": HiveonS19jPro,
        "ANTMINER S19": HiveonS19,
        "ANTMINER S19K PRO": HiveonS19kPro,
        "ANTMINER S19X88": HiveonS19NoPIC,
    },
    MinerTypes.MSKMINER: {
        None: MSKMiner,
        "S19-88": MSKMinerS19NoPIC,
    },
    MinerTypes.LUX_OS: {
        None: LUXMiner,
        "ANTMINER S9": LUXMinerS9,
        "ANTMINER S19": LUXMinerS19,
        "ANTMINER S19 PRO": LUXMinerS19Pro,
        "ANTMINER S19J PRO": LUXMinerS19jPro,
        "ANTMINER S19J PRO+": LUXMinerS19jProPlus,
        "ANTMINER S19K PRO": LUXMinerS19kPro,
        "ANTMINER S19 XP": LUXMinerS19XP,
        "ANTMINER T19": LUXMinerT19,
        "ANTMINER S21": LUXMinerS21,
    },
    MinerTypes.AURADINE: {
        None: type("AuradineUnknown", (Auradine, AuradineMake), {}),
        "AT1500": AuradineFluxAT1500,
        "AT2860": AuradineFluxAT2860,
        "AT2880": AuradineFluxAT2880,
        "AI2500": AuradineFluxAI2500,
        "AI3680": AuradineFluxAI3680,
        "AD2500": AuradineFluxAD2500,
        "AD3500": AuradineFluxAD3500,
    },
    MinerTypes.MARATHON: {
        None: MaraMiner,
        "ANTMINER S19": MaraS19,
        "ANTMINER S19 PRO": MaraS19Pro,
        "ANTMINER S19J": MaraS19j,
        "ANTMINER S19J88NOPIC": MaraS19jNoPIC,
        "ANTMINER S19J PRO": MaraS19jPro,
        "ANTMINER S19 XP": MaraS19XP,
        "ANTMINER S19K PRO": MaraS19KPro,
        "ANTMINER S21": MaraS21,
        "ANTMINER T21": MaraT21,
    },
    MinerTypes.BITAXE: {
        None: BitAxe,
        "BM1368": BitAxeSupra,
        "BM1366": BitAxeUltra,
        "BM1397": BitAxeMax,
        "BM1370": BitAxeGamma,
    },
    MinerTypes.LUCKYMINER: {
        None: LuckyMiner,
        "LV08": LuckyMinerLV08,
        "LV07": LuckyMinerLV07,
    },
    MinerTypes.ICERIVER: {
        None: type("IceRiverUnknown", (IceRiver, IceRiverMake), {}),
        "KS0": IceRiverKS0,
        "KS1": IceRiverKS1,
        "KS2": IceRiverKS2,
        "KS3": IceRiverKS3,
        "KS3L": IceRiverKS3L,
        "KS3M": IceRiverKS3M,
        "KS5": IceRiverKS5,
        "KS5L": IceRiverKS5L,
        "KS5M": IceRiverKS5M,
    },
    MinerTypes.HAMMER: {
        None: type("HammerUnknown", (BlackMiner, HammerMake), {}),
        "HAMMER D10": HammerD10,
    },
    MinerTypes.VOLCMINER: {
        None: type("VolcMinerUnknown", (BlackMiner, VolcMinerMake), {}),
        "VOLCMINER D1": VolcMinerD1,
    },
    MinerTypes.ELPHAPEX: {
        None: type("ElphapexUnknown", (ElphapexMiner, ElphapexMake), {}),
        "DG1+": ElphapexDG1Plus,
    },
}


async def concurrent_get_first_result(tasks: list, verification_func: Callable) -> Any:
    res = None
    for fut in asyncio.as_completed(tasks):
        res = await fut
        if verification_func(res):
            break
    for t in tasks:
        t.cancel()
        try:
            await t
        except asyncio.CancelledError:
            pass
    return res


class MinerFactory:
    async def get_multiple_miners(
        self, ips: list[str], limit: int = 200
    ) -> list[AnyMiner]:
        results = []

        async for miner in self.get_miner_generator(ips, limit):
            results.append(miner)

        return results

    async def get_miner_generator(
        self, ips: list, limit: int = 200
    ) -> AsyncGenerator[AnyMiner]:
        tasks = []
        semaphore = asyncio.Semaphore(limit)

        for ip in ips:
            tasks.append(asyncio.create_task(self.get_miner(ip)))

        for task in tasks:
            async with semaphore:
                result = await task
                if result is not None:
                    yield result

    async def get_miner(self, ip: str | ipaddress.ip_address) -> AnyMiner | None:
        ip = str(ip)

        miner_type = None

        for _ in range(settings.get("factory_get_retries", 1)):
            task = asyncio.create_task(self._get_miner_type(ip))
            try:
                miner_type = await asyncio.wait_for(
                    task, timeout=settings.get("factory_get_timeout", 3)
                )
            except asyncio.TimeoutError:
                continue
            else:
                if miner_type is not None:
                    break

        if miner_type is not None:
            miner_model = None
            miner_model_fns = {
                MinerTypes.ANTMINER: self.get_miner_model_antminer,
                MinerTypes.WHATSMINER: self.get_miner_model_whatsminer,
                MinerTypes.AVALONMINER: self.get_miner_model_avalonminer,
                MinerTypes.INNOSILICON: self.get_miner_model_innosilicon,
                MinerTypes.GOLDSHELL: self.get_miner_model_goldshell,
                MinerTypes.BRAIINS_OS: self.get_miner_model_braiins_os,
                MinerTypes.VNISH: self.get_miner_model_vnish,
                MinerTypes.EPIC: self.get_miner_model_epic,
                MinerTypes.HIVEON: self.get_miner_model_hiveon,
                MinerTypes.LUX_OS: self.get_miner_model_luxos,
                MinerTypes.AURADINE: self.get_miner_model_auradine,
                MinerTypes.MARATHON: self.get_miner_model_marathon,
                MinerTypes.BITAXE: self.get_miner_model_bitaxe,
                MinerTypes.LUCKYMINER: self.get_miner_model_luckyminer,
                MinerTypes.ICERIVER: self.get_miner_model_iceriver,
                MinerTypes.HAMMER: self.get_miner_model_hammer,
                MinerTypes.VOLCMINER: self.get_miner_model_volcminer,
                MinerTypes.ELPHAPEX: self.get_miner_model_elphapex,
            }
            fn = miner_model_fns.get(miner_type)

            if fn is not None:
                # noinspection PyArgumentList
                task = asyncio.create_task(fn(ip))
                try:
                    miner_model = await asyncio.wait_for(
                        task, timeout=settings.get("factory_get_timeout", 3)
                    )
                except asyncio.TimeoutError:
                    pass
            miner = self._select_miner_from_classes(
                ip,
                miner_type=miner_type,
                miner_model=miner_model,
            )
            return miner

    async def _get_miner_type(self, ip: str) -> MinerTypes | None:
        tasks = [
            asyncio.create_task(self._get_miner_web(ip)),
            asyncio.create_task(self._get_miner_socket(ip)),
        ]

        return await concurrent_get_first_result(tasks, lambda x: x is not None)

    async def _get_miner_web(self, ip: str) -> MinerTypes | None:
        urls = [f"http://{ip}/", f"https://{ip}/"]
        async with httpx.AsyncClient(
            transport=settings.transport(verify=False)
        ) as session:
            tasks = [asyncio.create_task(self._web_ping(session, url)) for url in urls]

            text, resp = await concurrent_get_first_result(
                tasks,
                lambda x: x[0] is not None
                and self._parse_web_type(x[0], x[1]) is not None,
            )
            if text is not None:
                mtype = self._parse_web_type(text, resp)
                if mtype == MinerTypes.ANTMINER:
                    # could still be mara
                    auth = httpx.DigestAuth("root", "root")
                    res = await self.send_web_command(ip, "/kaonsu/v1/brief", auth=auth)
                    if res is not None:
                        mtype = MinerTypes.MARATHON
                if mtype == MinerTypes.HAMMER:
                    res = await self.get_miner_model_hammer(ip)
                    if res is None:
                        return MinerTypes.HAMMER
                    if "HAMMER" in res.upper():
                        mtype = MinerTypes.HAMMER
                    else:
                        mtype = MinerTypes.VOLCMINER
                return mtype

    @staticmethod
    async def _web_ping(
        session: httpx.AsyncClient, url: str
    ) -> tuple[str | None, httpx.Response | None]:
        try:
            resp = await session.get(url, follow_redirects=True)
            return resp.text, resp
        except (
            httpx.HTTPError,
            asyncio.TimeoutError,
            anyio.EndOfStream,
            anyio.ClosedResourceError,
        ):
            pass
        return None, None

    @staticmethod
    def _parse_web_type(web_text: str, web_resp: httpx.Response) -> MinerTypes | None:
        if web_resp.status_code == 401 and 'realm="antMiner' in web_resp.headers.get(
            "www-authenticate", ""
        ):
            return MinerTypes.ANTMINER
        if web_resp.status_code == 401 and 'realm="blackMiner' in web_resp.headers.get(
            "www-authenticate", ""
        ):
            return MinerTypes.HAMMER
        if web_resp.status_code == 401 and 'realm="Daoge' in web_resp.headers.get(
            "www-authenticate", ""
        ):
            return MinerTypes.ELPHAPEX
        if len(web_resp.history) > 0:
            history_resp = web_resp.history[0]
            if (
                "/cgi-bin/luci" in web_text
                and history_resp.status_code == 307
                and "https://" in history_resp.headers.get("location", "")
            ):
                return MinerTypes.WHATSMINER
        if "Braiins OS" in web_text:
            return MinerTypes.BRAIINS_OS
        if "Luxor Firmware" in web_text:
            return MinerTypes.LUX_OS
        if "<TITLE>用户界面</TITLE>" in web_text:
            return MinerTypes.ICERIVER
        if "AxeOS" in web_text:
            return MinerTypes.BITAXE
        if "Lucky miner" in web_text:
            return MinerTypes.LUCKYMINER
        if "cloud-box" in web_text:
            return MinerTypes.GOLDSHELL
        if "AnthillOS" in web_text:
            return MinerTypes.VNISH
        if "Miner Web Dashboard" in web_text:
            return MinerTypes.EPIC
        if "Avalon" in web_text:
            return MinerTypes.AVALONMINER
        if "DragonMint" in web_text:
            return MinerTypes.INNOSILICON
        if "Miner UI" in web_text:
            return MinerTypes.AURADINE
        if "<title>Antminer</title>" in web_text:
            return MinerTypes.MSKMINER

    async def _get_miner_socket(self, ip: str) -> MinerTypes | None:
        commands = ["version", "devdetails"]
        tasks = [asyncio.create_task(self._socket_ping(ip, cmd)) for cmd in commands]

        data = await concurrent_get_first_result(
            tasks,
            lambda x: x is not None and self._parse_socket_type(x) is not None,
        )
        if data is not None:
            d = self._parse_socket_type(data)
            return d

    @staticmethod
    async def _socket_ping(ip: str, cmd: str) -> str | None:
        data = b""
        try:
            reader, writer = await asyncio.wait_for(
                asyncio.open_connection(str(ip), 4028),
                timeout=settings.get("factory_get_timeout", 3),
            )
        except (ConnectionError, OSError, asyncio.TimeoutError):
            return

        cmd = {"command": cmd}

        try:
            # send the command
            writer.write(json.dumps(cmd).encode("utf-8"))
            await writer.drain()

            # loop to receive all the data
            timeouts_remaining = max(1, int(settings.get("factory_get_timeout", 3)))
            while True:
                try:
                    d = await asyncio.wait_for(reader.read(4096), timeout=1)
                    if not d:
                        break
                    data += d
                except asyncio.TimeoutError:
                    timeouts_remaining -= 1
                    if not timeouts_remaining:
                        logger.warning(f"{ip}: Socket ping timeout.")
                        break
                except ConnectionResetError:
                    return
        except asyncio.CancelledError:
            raise
        except (ConnectionError, OSError):
            return
        finally:
            # Handle cancellation explicitly
            if writer.transport.is_closing():
                writer.transport.close()
            else:
                writer.close()
            try:
                await writer.wait_closed()
            except (ConnectionError, OSError):
                return
        if data:
            return data.decode("utf-8")

    @staticmethod
    def _parse_socket_type(data: str) -> MinerTypes | None:
        upper_data = data.upper()
        if "BOSMINER" in upper_data or "BOSER" in upper_data:
            return MinerTypes.BRAIINS_OS
        if "BTMINER" in upper_data or "BITMICRO" in upper_data:
            return MinerTypes.WHATSMINER
        if "LUXMINER" in upper_data:
            return MinerTypes.LUX_OS
        if "HIVEON" in upper_data:
            return MinerTypes.HIVEON
        if "KAONSU" in upper_data:
            return MinerTypes.MARATHON
        if "RWGLR" in upper_data:
            return MinerTypes.MSKMINER
        if "ANTMINER" in upper_data and "DEVDETAILS" not in upper_data:
            return MinerTypes.ANTMINER
        if (
            "INTCHAINS_QOMO" in upper_data
            or "KDAMINER" in upper_data
            or "BFGMINER" in upper_data
        ):
            return MinerTypes.GOLDSHELL
        if "INNOMINER" in upper_data:
            return MinerTypes.INNOSILICON
        if "AVALON" in upper_data:
            return MinerTypes.AVALONMINER
        if "GCMINER" in upper_data or "FLUXOS" in upper_data:
            return MinerTypes.AURADINE
        if "VNISH" in upper_data:
            return MinerTypes.VNISH

    async def send_web_command(
        self,
        ip: str,
        location: str,
        auth: httpx.DigestAuth = None,
    ) -> dict | None:
        async with httpx.AsyncClient(transport=settings.transport()) as session:
            try:
                data = await session.get(
                    f"http://{ip}{location}",
                    auth=auth,
                    timeout=settings.get("factory_get_timeout", 3),
                )
            except (httpx.HTTPError, asyncio.TimeoutError):
                logger.info(f"{ip}: Web command timeout.")
                return
        if data is None:
            return
        try:
            json_data = data.json()
        except (json.JSONDecodeError, asyncio.TimeoutError):
            try:
                return json.loads(data.text)
            except (json.JSONDecodeError, httpx.HTTPError):
                return
        else:
            return json_data

    async def send_api_command(self, ip: str, command: str) -> dict | None:
        data = b""
        try:
            reader, writer = await asyncio.open_connection(ip, 4028)
        except (ConnectionError, OSError):
            return
        cmd = {"command": command}

        try:
            # send the command
            writer.write(json.dumps(cmd).encode("utf-8"))
            await writer.drain()

            # loop to receive all the data
            while True:
                d = await reader.read(4096)
                if not d:
                    break
                data += d

            writer.close()
            await writer.wait_closed()
        except asyncio.CancelledError:
            writer.close()
            await writer.wait_closed()
            return
        except (ConnectionError, OSError):
            return
        if data == b"Socket connect failed: Connection refused\n":
            return

        data = await self._fix_api_data(data)

        try:
            data = json.loads(data)
        except json.JSONDecodeError:
            return {}

        return data

    @staticmethod
    async def _fix_api_data(data: bytes) -> str:
        if data.endswith(b"\x00"):
            str_data = data.decode("utf-8")[:-1]
        else:
            str_data = data.decode("utf-8")
        # fix an error with a btminer return having an extra comma that breaks json.loads()
        str_data = str_data.replace(",}", "}")
        # fix an error with a btminer return having a newline that breaks json.loads()
        str_data = str_data.replace("\n", "")
        # fix an error with a bmminer return not having a specific comma that breaks json.loads()
        str_data = str_data.replace("}{", "},{")
        # fix an error with a bmminer return having a specific comma that breaks json.loads()
        str_data = str_data.replace("[,{", "[{")
        # fix an error with a btminer return having a missing comma. (2023-01-06 version)
        str_data = str_data.replace('""temp0', '","temp0')
        # fix an error with Avalonminers returning inf and nan
        str_data = str_data.replace('"inf"', "0")
        str_data = str_data.replace('"nan"', "0")
        # fix whatever this garbage from avalonminers is `,"id":1}`
        if str_data.startswith(","):
            str_data = f"{{{str_data[1:]}"
        # try to fix an error with overflowing the recieve buffer
        # this can happen in cases such as bugged btminers returning arbitrary length error info with 100s of errors.
        if not str_data.endswith("}"):
            str_data = ",".join(str_data.split(",")[:-1]) + "}"

        # fix a really nasty bug with whatsminer API v2.0.4 where they return a list structured like a dict
        if re.search(r"\"error_code\":\[\".+\"]", str_data):
            str_data = str_data.replace("[", "{").replace("]", "}")

        return str_data

    @staticmethod
    def _select_miner_from_classes(
        ip: ipaddress.ip_address,
        miner_model: str | None,
        miner_type: MinerTypes | None,
    ) -> AnyMiner | None:
        # special case since hiveon miners return web results copying the antminer stock FW
        if "HIVEON" in str(miner_model).upper():
            miner_model = str(miner_model).upper().replace(" HIVEON", "")
            miner_type = MinerTypes.HIVEON
        try:
            return MINER_CLASSES[miner_type][str(miner_model).upper()](ip)
        except LookupError:
            if miner_type in MINER_CLASSES:
                if miner_model is not None:
                    warnings.warn(
                        f"Partially supported miner found: {miner_model}, type: {miner_type}, please open an issue with miner data "
                        f"and this model on GitHub (https://github.com/UpstreamData/pyasic/issues)."
                    )
                return MINER_CLASSES[miner_type][None](ip)
            return UnknownMiner(str(ip))

    async def get_miner_model_antminer(self, ip: str) -> str | None:
        tasks = [
            asyncio.create_task(self._get_model_antminer_web(ip)),
            asyncio.create_task(self._get_model_antminer_sock(ip)),
        ]

        return await concurrent_get_first_result(tasks, lambda x: x is not None)

    async def _get_model_antminer_web(self, ip: str) -> str | None:
        # last resort, this is slow
        auth = httpx.DigestAuth(
            "root", settings.get("default_antminer_web_password", "root")
        )
        web_json_data = await self.send_web_command(
            ip, "/cgi-bin/get_system_info.cgi", auth=auth
        )

        try:
            miner_model = web_json_data["minertype"]

            return miner_model
        except (TypeError, LookupError):
            pass

    async def _get_model_antminer_sock(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "version")
        try:
            miner_model = sock_json_data["VERSION"][0]["Type"]

            if " (" in miner_model:
                split_miner_model = miner_model.split(" (")
                miner_model = split_miner_model[0]

            return miner_model
        except (TypeError, LookupError):
            pass

        sock_json_data = await self.send_api_command(ip, "stats")
        try:
            miner_model = sock_json_data["STATS"][0]["Type"]

            if " (" in miner_model:
                split_miner_model = miner_model.split(" (")
                miner_model = split_miner_model[0]

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_goldshell(self, ip: str) -> str | None:
        json_data = await self.send_web_command(ip, "/mcb/status")

        try:
            miner_model = json_data["model"].replace("-", " ")

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_whatsminer(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "devdetails")
        try:
            miner_model = sock_json_data["DEVDETAILS"][0]["Model"].replace("_", "")
            miner_model = miner_model[:-1] + "0"

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_avalonminer(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "version")
        try:
            miner_model = sock_json_data["VERSION"][0]["PROD"].upper()
            if "-" in miner_model:
                miner_model = miner_model.split("-")[0]
            if miner_model in ["AVALONNANO", "AVALON0O", "AVALONMINER 15"]:
                subtype = sock_json_data["VERSION"][0]["MODEL"].upper()
                miner_model = f"AVALONMINER {subtype}"
            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_innosilicon(self, ip: str) -> str | None:
        try:
            async with httpx.AsyncClient(transport=settings.transport()) as session:
                auth_req = await session.post(
                    f"http://{ip}/api/auth",
                    data={
                        "username": "admin",
                        "password": settings.get(
                            "default_innosilicon_web_password", "admin"
                        ),
                    },
                )
                auth = auth_req.json()["jwt"]
        except (httpx.HTTPError, LookupError):
            return

        try:
            async with httpx.AsyncClient(transport=settings.transport()) as session:
                web_data = (
                    await session.post(
                        f"http://{ip}/api/type",
                        headers={"Authorization": "Bearer " + auth},
                        data={},
                    )
                ).json()
                return web_data["type"]
        except (httpx.HTTPError, LookupError):
            pass
        try:
            async with httpx.AsyncClient(transport=settings.transport()) as session:
                web_data = (
                    await session.post(
                        f"http://{ip}/overview",
                        headers={"Authorization": "Bearer " + auth},
                        data={},
                    )
                ).json()
                return web_data["type"]
        except (httpx.HTTPError, LookupError):
            pass

    async def get_miner_model_braiins_os(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "devdetails")
        try:
            miner_model = (
                sock_json_data["DEVDETAILS"][0]["Model"]
                .replace("Bitmain ", "")
                .replace("S19XP", "S19 XP")
            )
            return miner_model
        except (TypeError, LookupError):
            pass

        try:
            async with httpx.AsyncClient(transport=settings.transport()) as session:
                d = await session.post(
                    f"http://{ip}/graphql",
                    json={"query": "{bosminer {info{modelName}}}"},
                )
            if d.status_code == 200:
                json_data = d.json()
                miner_model = json_data["data"]["bosminer"]["info"][
                    "modelName"
                ].replace("S19XP", "S19 XP")
                return miner_model
        except (httpx.HTTPError, LookupError):
            pass

    async def get_miner_model_vnish(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "stats")
        try:
            miner_model = sock_json_data["STATS"][0]["Type"]
            if " (" in miner_model:
                split_miner_model = miner_model.split(" (")
                miner_model = split_miner_model[0]

            if "(88)" in miner_model:
                miner_model = miner_model.replace("(88)", "NOPIC")

            if " AML" in miner_model:
                miner_model = miner_model.replace(" AML", "")

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_epic(self, ip: str) -> str | None:
        for retry_cnt in range(settings.get("get_data_retries", 1)):
            sock_json_data = await self.send_web_command(ip, ":4028/capabilities")
            try:
                miner_model = sock_json_data["Model"]
                return miner_model
            except (TypeError, LookupError):
                if retry_cnt < settings.get("get_data_retries", 1) - 1:
                    continue
                else:
                    pass

    async def get_miner_model_hiveon(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "version")
        try:
            miner_type = sock_json_data["VERSION"][0]["Type"]

            return miner_type.replace(" HIVEON", "")
        except (TypeError, LookupError):
            pass

    async def get_miner_model_luxos(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "version")
        try:
            miner_model = sock_json_data["VERSION"][0]["Type"]

            if " (" in miner_model:
                split_miner_model = miner_model.split(" (")
                miner_model = split_miner_model[0]
            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_auradine(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "devdetails")
        try:
            return sock_json_data["DEVDETAILS"][0]["Model"]
        except LookupError:
            pass

    async def get_miner_model_marathon(self, ip: str) -> str | None:
        auth = httpx.DigestAuth("root", "root")
        web_json_data = await self.send_web_command(
            ip, "/kaonsu/v1/overview", auth=auth
        )

        try:
            miner_model = web_json_data["model"]
            if miner_model == "":
                return None

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_bitaxe(self, ip: str) -> str | None:
        web_json_data = await self.send_web_command(ip, "/api/system/info")

        try:
            miner_model = web_json_data["ASICModel"]
            if miner_model == "":
                return None

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_luckyminer(self, ip: str) -> str | None:
        web_json_data = await self.send_web_command(ip, "/api/system/info")

        try:
            miner_model = web_json_data["minerModel"]
            if miner_model == "":
                return None

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_iceriver(self, ip: str) -> str | None:
        async with httpx.AsyncClient(transport=settings.transport()) as client:
            try:
                # auth
                await client.post(
                    f"http://{ip}/user/loginpost",
                    params={
                        "post": "6",
                        "user": "admin",
                        "pwd": settings.get(
                            "default_iceriver_web_password", "12345678"
                        ),
                    },
                )
            except httpx.HTTPError:
                return None
            try:
                resp = await client.post(
                    f"http://{ip}:/user/userpanel", params={"post": "4"}
                )
                if not resp.status_code == 200:
                    return
                result = resp.json()
                software_ver = result["data"]["softver1"]
                split_ver = software_ver.split("_")
                if split_ver[-1] == "miner":
                    miner_ver = split_ver[-2]
                else:
                    miner_ver = split_ver[-1].replace("miner", "")
                return miner_ver.upper()
            except httpx.HTTPError:
                pass

    async def get_miner_model_hammer(self, ip: str) -> str | None:
        auth = httpx.DigestAuth(
            "root", settings.get("default_hammer_web_password", "root")
        )
        web_json_data = await self.send_web_command(
            ip, "/cgi-bin/get_system_info.cgi", auth=auth
        )

        try:
            miner_model = web_json_data["minertype"]

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_volcminer(self, ip: str) -> str | None:
        auth = httpx.DigestAuth(
            "root", settings.get("default_volcminer_web_password", "root")
        )
        web_json_data = await self.send_web_command(
            ip, "/cgi-bin/get_system_info.cgi", auth=auth
        )

        try:
            miner_model = web_json_data["minertype"]

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_elphapex(self, ip: str) -> str | None:
        auth = httpx.DigestAuth(
            "root", settings.get("default_elphapex_web_password", "root")
        )
        web_json_data = await self.send_web_command(
            ip, "/cgi-bin/get_system_info.cgi", auth=auth
        )

        try:
            miner_model = web_json_data["minertype"]

            return miner_model
        except (TypeError, LookupError):
            pass

    async def get_miner_model_mskminer(self, ip: str) -> str | None:
        sock_json_data = await self.send_api_command(ip, "version")
        try:
            return sock_json_data["VERSION"][0]["Type"].split(" ")[0]
        except LookupError:
            pass


miner_factory = MinerFactory()


# abstracted version of get miner that is easier to access
async def get_miner(ip: ipaddress.ip_address | str) -> AnyMiner:
    return await miner_factory.get_miner(ip)



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M50SPlusPlusVK10


class BTMinerM50SPlusPlusVK10(M5X, M50SPlusPlusVK10):
    pass


from pyasic.miners.device.models import M50SPlusPlusVK20


class BTMinerM50SPlusPlusVK20(M5X, M50SPlusPlusVK20):
    pass


from pyasic.miners.device.models import M50SPlusPlusVK30


class BTMinerM50SPlusPlusVK30(M5X, M50SPlusPlusVK30):
    pass


from pyasic.miners.device.models import M50SPlusPlusVK40


class BTMinerM50SPlusPlusVK40(M5X, M50SPlusPlusVK40):
    pass


from pyasic.miners.device.models import M50SPlusPlusVK50


class BTMinerM50SPlusPlusVK50(M5X, M50SPlusPlusVK50):
    pass


from pyasic.miners.device.models import M50SPlusPlusVK60


class BTMinerM50SPlusPlusVK60(M5X, M50SPlusPlusVK60):
    pass


from pyasic.miners.device.models import M50SPlusPlusVL20


class BTMinerM50SPlusPlusVL20(M5X, M50SPlusPlusVL20):
    pass


from pyasic.miners.device.models import M50SPlusPlusVL30


class BTMinerM50SPlusPlusVL30(M5X, M50SPlusPlusVL30):
    pass


from pyasic.miners.device.models import M50SPlusPlusVL40


class BTMinerM50SPlusPlusVL40(M5X, M50SPlusPlusVL40):
    pass


from pyasic.miners.device.models import M50SPlusPlusVL50


class BTMinerM50SPlusPlusVL50(M5X, M50SPlusPlusVL50):
    pass


from pyasic.miners.device.models import M50SPlusPlusVL60


class BTMinerM50SPlusPlusVL60(M5X, M50SPlusPlusVL60):
    pass



from pyasic.miners.backends import M3X
from pyasic.miners.device.models import M31SV10


class BTMinerM31SV10(M3X, M31SV10):
    pass


from pyasic.miners.device.models import M31SV20


class BTMinerM31SV20(M3X, M31SV20):
    pass


from pyasic.miners.device.models import M31SV30


class BTMinerM31SV30(M3X, M31SV30):
    pass


from pyasic.miners.device.models import M31SV40


class BTMinerM31SV40(M3X, M31SV40):
    pass


from pyasic.miners.device.models import M31SV50


class BTMinerM31SV50(M3X, M31SV50):
    pass


from pyasic.miners.device.models import M31SV60


class BTMinerM31SV60(M3X, M31SV60):
    pass


from pyasic.miners.device.models import M31SV70


class BTMinerM31SV70(M3X, M31SV70):
    pass


from pyasic.miners.device.models import M31SV80


class BTMinerM31SV80(M3X, M31SV80):
    pass


from pyasic.miners.device.models import M31SV90


class BTMinerM31SV90(M3X, M31SV90):
    pass


from pyasic.miners.device.models import M31SVE10


class BTMinerM31SVE10(M3X, M31SVE10):
    pass


from pyasic.miners.device.models import M31SVE20


class BTMinerM31SVE20(M3X, M31SVE20):
    pass


from pyasic.miners.device.models import M31SVE30


class BTMinerM31SVE30(M3X, M31SVE30):
    pass



from pyasic import settings
from pyasic.ssh.base import BaseSSH


class BOSMinerSSH(BaseSSH):
    def __init__(self, ip: str):
        """
        Initialize a BOSMinerSSH instance.

        Args:
            ip (str): The IP address of the BOSMiner device.
        """
        super().__init__(ip)
        self.pwd = settings.get("default_bosminer_ssh_password", "root")

    async def get_board_info(self):
        """
        Retrieve information about the BOSMiner board.

        Returns:
            str: Information about the BOSMiner board.
        """
        return await self.send_command("bosminer model -d")

    async def fault_light_on(self):
        """
        Turn on the fault light of the BOSMiner device.

        Returns:
            str: Confirmation message after turning on the fault light.
        """
        return await self.send_command("miner fault_light on")

    async def fault_light_off(self):
        """
        Turn off the fault light of the BOSMiner device.

        Returns:
            str: Confirmation message after turning off the fault light.
        """
        return await self.send_command("miner fault_light off")

    async def restart_bosminer(self):
        """
        Restart the BOSMiner service on the device.

        Returns:
            str: Confirmation message after restarting the BOSMiner service.
        """
        return await self.send_command("/etc/init.d/bosminer restart")

    async def reboot(self):
        """
        Reboot the BOSMiner device.

        Returns:
            str: Confirmation message after initiating the reboot process.
        """
        return await self.send_command("/sbin/reboot")

    async def get_config_file(self):
        """
        Retrieve the configuration file of BOSMiner.

        Returns:
            str: Content of the BOSMiner configuration file.
        """
        return await self.send_command("cat /etc/bosminer.toml")

    async def get_network_config(self):
        """
        Retrieve the network configuration of the BOSMiner device.

        Returns:
            str: Content of the network configuration file.
        """
        return await self.send_command("cat /etc/config/network")

    async def get_hostname(self):
        """
        Retrieve the hostname of the BOSMiner device.

        Returns:
            str: Hostname of the BOSMiner device.
        """
        return await self.send_command("cat /proc/sys/kernel/hostname")

    async def get_led_status(self):
        """
        Retrieve the status of the LED on the BOSMiner device.

        Returns:
            str: Status of the LED.
        """
        return await self.send_command("cat /sys/class/leds/'Red LED'/delay_off")



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M31SPlusV10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV10

    expected_chips = 105
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV100(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV100

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV20

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV30

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV40

    expected_chips = 123
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV50

    expected_chips = 148
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV60

    expected_chips = 156
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV80

    expected_chips = 129
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusV90(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV90

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVE10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE10

    expected_chips = 82
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVE20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE20

    expected_chips = 78
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVE30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE30

    expected_chips = 105
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVE40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE40

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVE50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE50

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVE60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE60

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVE80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE80

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVF20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVF20

    expected_chips = 66
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVF30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVF30

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVG20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVG20

    expected_chips = 66
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SPlusVG30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVG30

    expected_chips = 70
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M61SVL10


class BTMinerM61SVL10(M6X, M61SVL10):
    pass


from pyasic.miners.device.models import M61SVL20


class BTMinerM61SVL20(M6X, M61SVL20):
    pass


from pyasic.miners.device.models import M61SVL30


class BTMinerM61SVL30(M6X, M61SVL30):
    pass



from pyasic.miners.backends import M3X
from pyasic.miners.device.models import M33SPlusVG20


class BTMinerM33SPlusVG20(M3X, M33SPlusVG20):
    pass


from pyasic.miners.device.models import M33SPlusVG30


class BTMinerM33SPlusVG30(M3X, M33SPlusVG30):
    pass


from pyasic.miners.device.models import M33SPlusVH20


class BTMinerM33SPlusVH20(M3X, M33SPlusVH20):
    pass


from pyasic.miners.device.models import M33SPlusVH30


class BTMinerM33SPlusVH30(M3X, M33SPlusVH30):
    pass



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import BitAxeMake


class Supra(BitAxeMake):
    raw_model = MinerModel.BITAXE.BM1368

    expected_hashboards = 1
    expected_chips = 1
    expected_fans = 1
    algo = MinerAlgo.SHA256



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M53SVH20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SVH20

    expected_chips = 198
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SVH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SVH30

    expected_chips = 204
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SVJ30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SVJ30

    expected_chips = 180
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SVJ40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SVJ40

    expected_chips = 192
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SVK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SVK30

    expected_chips = 128
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

import logging
from typing import List, Optional

from pyasic import MinerConfig
from pyasic.data.error_codes import MinerErrorData, VnishError
from pyasic.device.algorithm import AlgoHashRate
from pyasic.errors import APIError
from pyasic.miners.backends.bmminer import BMMiner
from pyasic.miners.data import (
    DataFunction,
    DataLocations,
    DataOptions,
    RPCAPICommand,
    WebAPICommand,
)
from pyasic.miners.device.firmware import VNishFirmware
from pyasic.web.vnish import VNishWebAPI

VNISH_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.WATTAGE_LIMIT): DataFunction(
            "_get_wattage_limit",
            [WebAPICommand("web_settings", "settings")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
        str(DataOptions.ERRORS): DataFunction(
            "_get_errors",
            [WebAPICommand("web_summary", "summary")],
        ),
    }
)


class VNish(VNishFirmware, BMMiner):
    """Handler for VNish miners"""

    _web_cls = VNishWebAPI
    web: VNishWebAPI

    supports_shutdown = True
    supports_presets = True
    supports_autotuning = True

    data_locations = VNISH_DATA_LOC

    async def send_config(self, config: MinerConfig, user_suffix: str = None) -> None:
        await self.web.post_settings(
            miner_settings=config.as_vnish(user_suffix=user_suffix)
        )

    async def restart_backend(self) -> bool:
        data = await self.web.restart_vnish()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def stop_mining(self) -> bool:
        data = await self.web.stop_mining()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def resume_mining(self) -> bool:
        data = await self.web.resume_mining()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def reboot(self) -> bool:
        data = await self.web.reboot()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def _get_mac(self, web_summary: dict = None) -> str:
        if web_summary is not None:
            try:
                mac = web_summary["system"]["network_status"]["mac"]
                return mac
            except KeyError:
                pass

        web_info = await self.web.info()

        if web_info is not None:
            try:
                mac = web_info["system"]["network_status"]["mac"]
                return mac
            except KeyError:
                pass

    async def fault_light_off(self) -> bool:
        result = await self.web.find_miner()
        if result is not None:
            if result.get("on") is False:
                return True
            else:
                await self.web.find_miner()

    async def fault_light_on(self) -> bool:
        result = await self.web.find_miner()
        if result is not None:
            if result.get("on") is True:
                return True
            else:
                await self.web.find_miner()

    async def _get_hostname(self, web_summary: dict = None) -> str:
        if web_summary is None:
            web_info = await self.web.info()

            if web_info is not None:
                try:
                    hostname = web_info["system"]["network_status"]["hostname"]
                    return hostname
                except KeyError:
                    pass

        if web_summary is not None:
            try:
                hostname = web_summary["system"]["network_status"]["hostname"]
                return hostname
            except KeyError:
                pass

    async def _get_wattage(self, web_summary: dict = None) -> Optional[int]:
        if web_summary is None:
            web_summary = await self.web.summary()

        if web_summary is not None:
            try:
                wattage = web_summary["miner"]["power_usage"]
                wattage = round(wattage)
                return wattage
            except KeyError:
                pass

    async def _get_hashrate(self, rpc_summary: dict = None) -> Optional[AlgoHashRate]:
        # get hr from API
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return self.algo.hashrate(
                    rate=float(rpc_summary["SUMMARY"][0]["GHS 5s"]),
                    unit=self.algo.unit.GH,
                ).into(self.algo.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_wattage_limit(self, web_settings: dict = None) -> Optional[int]:
        if web_settings is None:
            web_settings = await self.web.summary()

        if web_settings is not None:
            try:
                wattage_limit = web_settings["miner"]["overclock"]["preset"]
                if wattage_limit == "disabled":
                    return None
                return int(wattage_limit)
            except (KeyError, TypeError):
                pass

    async def _get_fw_ver(self, web_summary: dict = None) -> Optional[str]:
        if web_summary is None:
            web_summary = await self.web.summary()

        fw_ver = None
        if web_summary is not None:
            try:
                fw_ver = web_summary["miner"]["miner_type"]
                fw_ver = fw_ver.split("(Vnish ")[1].replace(")", "")
                return fw_ver
            except LookupError:
                return fw_ver

    async def _is_mining(self, web_summary: dict = None) -> Optional[bool]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                is_mining = not web_summary["miner"]["miner_status"]["miner_state"] in [
                    "stopped",
                    "shutting-down",
                    "failure",
                ]
                return is_mining
            except LookupError:
                pass

    async def _get_errors(self, web_summary: dict = None) -> List[MinerErrorData]:
        errors = []
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                return errors

        if web_summary is not None:
            chains = web_summary.get("miner", {}).get("chains", [])
            for chain in chains:
                state = chain.get("status", {}).get("state")
                description = chain.get("status", {}).get("description", "")
                if state == "failure":
                    errors.append(VnishError(error_message=description))

        return errors

    async def get_config(self) -> MinerConfig:
        try:
            web_settings = await self.web.settings()
            web_presets = await self.web.autotune_presets()
        except APIError:
            return self.config
        self.config = MinerConfig.from_vnish(web_settings, web_presets)
        return self.config

    async def set_power_limit(self, wattage: int) -> bool:
        config = await self.get_config()
        valid_presets = [
            preset.power
            for preset in config.mining_mode.available_presets
            if preset.tuned and preset.power <= wattage
        ]
        new_wattage = max(valid_presets)

        # Set power to highest preset <= wattage
        try:
            await self.web.set_power_limit(new_wattage)
            updated_settings = await self.web.settings()
        except APIError:
            raise
        except Exception as e:
            logging.warning(f"{self} - Failed to set power limit: {e}")
            return False

        if int(updated_settings["miner"]["overclock"]["preset"]) == new_wattage:
            return True
        else:
            return False



from pyasic.miners.backends import M3X
from pyasic.miners.device.models import M30V10


class BTMinerM30V10(M3X, M30V10):
    pass


from pyasic.miners.device.models import M30V20


class BTMinerM30V20(M3X, M30V20):
    pass



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M60SPlusVK30


class BTMinerM60SPlusVK30(M6X, M60SPlusVK30):
    pass


from pyasic.miners.device.models import M60SPlusVK40


class BTMinerM60SPlusVK40(M6X, M60SPlusVK40):
    pass


from pyasic.miners.device.models import M60SPlusVK50


class BTMinerM60SPlusVK50(M6X, M60SPlusVK50):
    pass


from pyasic.miners.device.models import M60SPlusVK60


class BTMinerM60SPlusVK60(M6X, M60SPlusVK60):
    pass


from pyasic.miners.device.models import M60SPlusVK70


class BTMinerM60SPlusVK70(M6X, M60SPlusVK70):
    pass


from pyasic.miners.device.models import M60SPlusVL10


class BTMinerM60SPlusVL10(M6X, M60SPlusVL10):
    pass


from pyasic.miners.device.models import M60SPlusVL30


class BTMinerM60SPlusVL30(M6X, M60SPlusVL30):
    pass


from pyasic.miners.device.models import M60SPlusVL40


class BTMinerM60SPlusVL40(M6X, M60SPlusVL40):
    pass


from pyasic.miners.device.models import M60SPlusVL50


class BTMinerM60SPlusVL50(M6X, M60SPlusVL50):
    pass


from pyasic.miners.device.models import M60SPlusVL60


class BTMinerM60SPlusVL60(M6X, M60SPlusVL60):
    pass



from .blake256 import Blake256Unit
from .eaglesong import EaglesongUnit
from .equihash import EquihashUnit
from .ethash import EtHashUnit
from .handshake import HandshakeUnit
from .kadena import KadenaUnit
from .kheavyhash import KHeavyHashUnit
from .scrypt import ScryptUnit
from .sha256 import SHA256Unit
from .x11 import X11Unit


class HashUnit:
    SHA256 = SHA256Unit
    SCRYPT = ScryptUnit
    KHEAVYHASH = KHeavyHashUnit
    KADENA = KadenaUnit
    HANDSHAKE = HandshakeUnit
    X11 = X11Unit
    BLAKE256 = Blake256Unit
    EAGLESONG = EaglesongUnit
    ETHASH = EtHashUnit
    EQUIHASH = EquihashUnit



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

from pyasic.miners.backends import HiveonModern
from pyasic.miners.device.models import (
    S19,
    S19L,
    S19XP,
    S19a,
    S19aPro,
    S19Hydro,
    S19i,
    S19j,
    S19jNoPIC,
    S19jPro,
    S19KPro,
    S19NoPIC,
    S19Plus,
    S19Pro,
    S19ProHydro,
    S19ProPlus,
    S19ProPlusHydro,
)


class HiveonS19(HiveonModern, S19):
    pass


class HiveonS19NoPIC(HiveonModern, S19NoPIC):
    pass


class HiveonS19Plus(HiveonModern, S19Plus):
    pass


class HiveonS19i(HiveonModern, S19i):
    pass


class HiveonS19Pro(HiveonModern, S19Pro):
    pass


class HiveonS19ProPlus(HiveonModern, S19ProPlus):
    pass


class HiveonS19XP(HiveonModern, S19XP):
    pass


class HiveonS19a(HiveonModern, S19a):
    pass


class HiveonS19aPro(HiveonModern, S19aPro):
    pass


class HiveonS19j(HiveonModern, S19j):
    pass


class HiveonS19jNoPIC(HiveonModern, S19jNoPIC):
    pass


class HiveonS19jPro(HiveonModern, S19jPro):
    pass


class HiveonS19L(HiveonModern, S19L):
    pass


class HiveonS19ProHydro(HiveonModern, S19ProHydro):
    pass


class HiveonS19Hydro(HiveonModern, S19Hydro):
    pass


class HiveonS19ProPlusHydro(HiveonModern, S19ProPlusHydro):
    pass


class HiveonS19kPro(HiveonModern, S19KPro):
    pass



from pyasic.device.algorithm.hashrate.unit.scrypt import ScryptUnit
from pyasic.device.algorithm.scrypt import ScryptHashRate
from pyasic.miners.backends import BlackMiner
from pyasic.miners.device.models import D10


class HammerD10(BlackMiner, D10):
    sticker_hashrate = ScryptHashRate(rate=5.0, unit=ScryptUnit.GH)



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import TypeVar, Union

from pydantic import Field

from pyasic.config.base import MinerConfigOption, MinerConfigValue


class FanModeNormal(MinerConfigValue):
    mode: str = Field(init=False, default="normal")
    minimum_fans: int = 1
    minimum_speed: int = 0

    @classmethod
    def from_dict(cls, dict_conf: dict | None) -> "FanModeNormal":
        cls_conf = {}
        if dict_conf.get("minimum_fans") is not None:
            cls_conf["minimum_fans"] = dict_conf["minimum_fans"]
        if dict_conf.get("minimum_speed") is not None:
            cls_conf["minimum_speed"] = dict_conf["minimum_speed"]
        return cls(**cls_conf)

    @classmethod
    def from_vnish(cls, web_cooling_settings: dict) -> "FanModeNormal":
        cls_conf = {}
        if web_cooling_settings.get("fan_min_count") is not None:
            cls_conf["minimum_fans"] = web_cooling_settings["fan_min_count"]
        if web_cooling_settings.get("fan_min_duty") is not None:
            cls_conf["minimum_speed"] = web_cooling_settings["fan_min_duty"]
        return cls(**cls_conf)

    @classmethod
    def from_bosminer(cls, toml_fan_conf: dict):
        cls_conf = {}
        if toml_fan_conf.get("min_fans") is not None:
            cls_conf["minimum_fans"] = toml_fan_conf["min_fans"]
        return cls(**cls_conf)

    def as_am_modern(self) -> dict:
        return {"bitmain-fan-ctrl": False, "bitmain-fan-pwn": "100"}

    def as_hiveon_modern(self) -> dict:
        return {"bitmain-fan-ctrl": False, "bitmain-fan-pwn": "100"}

    def as_elphapex(self) -> dict:
        return {"fc-fan-ctrl": False, "fc-fan-pwn": "100"}

    def as_bosminer(self) -> dict:
        return {
            "temp_control": {"mode": "auto"},
            "fan_control": {"min_fans": self.minimum_fans},
        }

    def as_epic(self) -> dict:
        return {
            "fans": {
                "Auto": {
                    "Idle Speed": (
                        self.minimum_speed if not self.minimum_speed == 0 else 100
                    )
                }
            }
        }

    def as_mara(self) -> dict:
        return {
            "general-config": {"environment-profile": "AirCooling"},
            "advance-config": {
                "override-fan-control": False,
                "fan-fixed-percent": 0,
            },
        }

    def as_espminer(self) -> dict:
        return {"autoFanspeed": 1}

    def as_luxos(self) -> dict:
        return {"fanset": {"speed": -1, "min_fans": self.minimum_fans}}

    def as_vnish(self) -> dict:
        return {
            "cooling": {
                "fan_min_count": self.minimum_fans,
                "fan_min_duty": self.minimum_speed,
                "mode": {
                    "name": "auto",
                    "param": None,  # Target temp, must be set later...
                },
            }
        }


class FanModeManual(MinerConfigValue):
    mode: str = Field(init=False, default="manual")
    speed: int = 100
    minimum_fans: int = 1

    @classmethod
    def from_dict(cls, dict_conf: dict | None) -> "FanModeManual":
        cls_conf = {}
        if dict_conf.get("speed") is not None:
            cls_conf["speed"] = dict_conf["speed"]
        if dict_conf.get("minimum_fans") is not None:
            cls_conf["minimum_fans"] = dict_conf["minimum_fans"]
        return cls(**cls_conf)

    @classmethod
    def from_bosminer(cls, toml_fan_conf: dict) -> "FanModeManual":
        cls_conf = {}
        if toml_fan_conf.get("min_fans") is not None:
            cls_conf["minimum_fans"] = toml_fan_conf["min_fans"]
        if toml_fan_conf.get("speed") is not None:
            cls_conf["speed"] = toml_fan_conf["speed"]
        return cls(**cls_conf)

    @classmethod
    def from_vnish(cls, web_cooling_settings: dict) -> "FanModeManual":
        cls_conf = {}
        if web_cooling_settings.get("fan_min_count") is not None:
            cls_conf["minimum_fans"] = web_cooling_settings["fan_min_count"]
        if web_cooling_settings["mode"].get("param") is not None:
            cls_conf["speed"] = web_cooling_settings["mode"]["param"]
        return cls(**cls_conf)

    def as_am_modern(self) -> dict:
        return {"bitmain-fan-ctrl": True, "bitmain-fan-pwm": str(self.speed)}

    def as_hiveon_modern(self) -> dict:
        return {"bitmain-fan-ctrl": True, "bitmain-fan-pwm": str(self.speed)}

    def as_elphapex(self) -> dict:
        return {"fc-fan-ctrl": True, "fc-fan-pwm": str(self.speed)}

    def as_bosminer(self) -> dict:
        return {
            "temp_control": {"mode": "manual"},
            "fan_control": {"min_fans": self.minimum_fans, "speed": self.speed},
        }

    def as_auradine(self) -> dict:
        return {"fan": {"percentage": self.speed}}

    def as_epic(self) -> dict:
        return {"fans": {"Manual": {"speed": self.speed}}}

    def as_mara(self) -> dict:
        return {
            "general-config": {"environment-profile": "AirCooling"},
            "advance-config": {
                "override-fan-control": True,
                "fan-fixed-percent": self.speed,
            },
        }

    def as_espminer(self) -> dict:
        return {"autoFanspeed": 0, "fanspeed": self.speed}

    def as_luxos(self) -> dict:
        return {"fanset": {"speed": self.speed, "min_fans": self.minimum_fans}}

    def as_vnish(self) -> dict:
        return {
            "cooling": {
                "fan_min_count": self.minimum_fans,
                "fan_min_duty": self.speed,
                "mode": {
                    "name": "manual",
                    "param": self.speed,  # Speed value
                },
            }
        }


class FanModeImmersion(MinerConfigValue):
    mode: str = Field(init=False, default="immersion")

    @classmethod
    def from_dict(cls, dict_conf: dict | None) -> "FanModeImmersion":
        return cls()

    def as_am_modern(self) -> dict:
        return {"bitmain-fan-ctrl": True, "bitmain-fan-pwm": "0"}

    def as_hiveon_modern(self) -> dict:
        return {"bitmain-fan-ctrl": True, "bitmain-fan-pwm": "0"}

    def as_elphapex(self) -> dict:
        return {"fc-fan-ctrl": True, "fc-fan-pwm": "0"}

    def as_bosminer(self) -> dict:
        return {
            "fan_control": {"min_fans": 0},
        }

    def as_auradine(self) -> dict:
        return {"fan": {"percentage": 0}}

    def as_mara(self) -> dict:
        return {"general-config": {"environment-profile": "OilImmersionCooling"}}

    def as_luxos(self) -> dict:
        return {"fanset": {"speed": 0, "min_fans": 0}}

    def as_vnish(self) -> dict:
        return {"cooling": {"mode": {"name": "immers"}}}


class FanModeConfig(MinerConfigOption):
    normal = FanModeNormal
    manual = FanModeManual
    immersion = FanModeImmersion

    @classmethod
    def default(cls):
        return cls.normal()

    @classmethod
    def from_dict(cls, dict_conf: dict | None):
        if dict_conf is None:
            return cls.default()

        mode = dict_conf.get("mode")
        if mode is None:
            return cls.default()

        cls_attr = getattr(cls, mode)
        if cls_attr is not None:
            return cls_attr().from_dict(dict_conf)

    @classmethod
    def from_am_modern(cls, web_conf: dict):
        if web_conf.get("bitmain-fan-ctrl") is not None:
            fan_manual = web_conf["bitmain-fan-ctrl"]
            if fan_manual:
                speed = int(web_conf["bitmain-fan-pwm"])
                if speed == 0:
                    return cls.immersion()
                return cls.manual(speed=speed)
            else:
                return cls.normal()
        else:
            return cls.default()

    @classmethod
    def from_hiveon_modern(cls, web_conf: dict):
        if web_conf.get("bitmain-fan-ctrl") is not None:
            fan_manual = web_conf["bitmain-fan-ctrl"]
            if fan_manual:
                speed = int(web_conf["bitmain-fan-pwm"])
                if speed == 0:
                    return cls.immersion()
                return cls.manual(speed=speed)
            else:
                return cls.normal()
        else:
            return cls.default()

    @classmethod
    def from_elphapex(cls, web_conf: dict):
        if web_conf.get("fc-fan-ctrl") is not None:
            fan_manual = web_conf["fc-fan-ctrl"]
            if fan_manual:
                speed = int(web_conf["fc-fan-pwm"])
                if speed == 0:
                    return cls.immersion()
                return cls.manual(speed=speed)
            else:
                return cls.normal()
        else:
            return cls.default()

    @classmethod
    def from_epic(cls, web_conf: dict):
        try:
            fan_mode = web_conf["Fans"]["Fan Mode"]
            if fan_mode.get("Manual") is not None:
                return cls.manual(speed=fan_mode.get("Manual"))
            else:
                return cls.normal()
        except KeyError:
            return cls.default()

    @classmethod
    def from_bosminer(cls, toml_conf: dict):
        try:
            mode = toml_conf["temp_control"]["mode"]
            fan_config = toml_conf.get("fan_control", {})
            if mode == "auto":
                return cls.normal().from_bosminer(fan_config)
            elif mode == "manual":
                if toml_conf.get("fan_control"):
                    return cls.manual().from_bosminer(fan_config)
                return cls.manual()
            elif mode == "disabled":
                return cls.immersion()
        except KeyError:
            pass

        try:
            min_fans = toml_conf["fan_control"]["min_fans"]
        except KeyError:
            return cls.default()

        if min_fans == 0:
            return cls.immersion()
        return cls.normal(minimum_fans=min_fans)

    @classmethod
    def from_vnish(cls, web_settings: dict):
        try:
            mode = web_settings["miner"]["cooling"]["mode"]["name"]
        except LookupError:
            return cls.default()

        if mode == "auto":
            return cls.normal().from_vnish(web_settings["miner"]["cooling"])
        elif mode == "manual":
            return cls.manual().from_vnish(web_settings["miner"]["cooling"])
        elif mode == "immers":
            return cls.immersion()

    @classmethod
    def from_boser(cls, grpc_miner_conf: dict):
        try:
            temperature_conf = grpc_miner_conf["temperature"]
        except LookupError:
            return cls.default()

        keys = temperature_conf.keys()
        if "auto" in keys:
            if "minimumRequiredFans" in keys:
                return cls.normal(minimum_fans=temperature_conf["minimumRequiredFans"])
            return cls.normal()
        if "manual" in keys:
            conf = {}
            if "fanSpeedRatio" in temperature_conf["manual"].keys():
                conf["speed"] = int(temperature_conf["manual"]["fanSpeedRatio"])
            if "minimumRequiredFans" in keys:
                conf["minimum_fans"] = int(temperature_conf["minimumRequiredFans"])
            return cls.manual(**conf)
        if "disabled" in keys:
            conf = {}
            if "fanSpeedRatio" in temperature_conf["disabled"].keys():
                conf["speed"] = int(temperature_conf["disabled"]["fanSpeedRatio"])
            return cls.manual(**conf)
        return cls.default()

    @classmethod
    def from_auradine(cls, web_fan: dict):
        try:
            fan_data = web_fan["Fan"][0]
            fan_1_max = fan_data["Max"]
            fan_1_target = fan_data["Target"]
            return cls.manual(speed=round((fan_1_target / fan_1_max) * 100))
        except LookupError:
            pass
        return cls.default()

    @classmethod
    def from_mara(cls, web_config: dict):
        try:
            mode = web_config["general-config"]["environment-profile"]
            if mode == "AirCooling":
                if web_config["advance-config"]["override-fan-control"]:
                    return cls.manual(
                        speed=web_config["advance-config"]["fan-fixed-percent"]
                    )
                return cls.normal()
            return cls.immersion()
        except LookupError:
            pass
        return cls.default()

    @classmethod
    def from_espminer(cls, web_system_info: dict):
        if web_system_info["autofanspeed"] == 1:
            return cls.normal()
        else:
            return cls.manual(speed=web_system_info["fanspeed"])

    @classmethod
    def from_luxos(cls, rpc_fans: dict, rpc_tempctrl: dict):
        try:
            mode = rpc_tempctrl["TEMPCTRL"][0]["Mode"]
            if mode == "Manual":
                speed = rpc_fans["FANS"][0]["Speed"]
                min_fans = rpc_fans["FANCTRL"][0]["MinFans"]
                if min_fans == 0 and speed == 0:
                    return cls.immersion()
                return cls.manual(
                    speed=speed,
                    minimum_fans=min_fans,
                )
            return cls.normal(
                minimum_fans=rpc_fans["FANCTRL"][0]["MinFans"],
            )
        except LookupError:
            pass
        return cls.default()


FanMode = TypeVar(
    "FanMode",
    bound=Union[FanModeNormal, FanModeManual, FanModeImmersion],
)
